import numpy as np
from pyspedas import clip, options, store_data, ylim, zlim, get_data

from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def hep(
    trange: List[str] = ['2017-03-27', '2017-03-28'],
    datatype: str = 'omniflux',
    level: str = 'l2',
    suffix: str = '',
    get_support_data: bool = True,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    version: Optional[str] = None,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the HEP experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-03-27', '2017-03-28']

        datatype: str
            Data type; Valid 'l2' options: 'omniflux', '3dflux'   Valid 'l3' options: 'pa'

        level: str
            Data level; Valid options: 'l2','l3'
            Default: 'l2'

        suffix: str
            The tplot variable names will be given this suffix.  Default: ''

        get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: True

        varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

        varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

        no_update: bool
            If set, only load data from your local cache. Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

        ror: bool
            If set, print PI info and rules of the road. Default: True

        version: str
            Set this value to specify the version of cdf files (such as "v01_02", "v01_03", ...)
            Default: None

        uname: str
            User name.  Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False



    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> hep_vars = pyspedas.projects.erg.hep(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_hep_l2_FEDO_L')


    """

    file_res = 3600. * 24
    prefix = 'erg_hep_'+level+'_'

    if level == 'l2':
        pathformat = 'satellite/erg/hep/'+level+'/'+datatype + \
            '/%Y/%m/erg_hep_'+level+'_'+datatype + '_%Y%m%d_'
        if version is None:
            pathformat += 'v??_??.cdf'
        else:
            pathformat += version + '.cdf'
    if level == 'l3':
        pathformat = 'satellite/erg/hep/'+level + \
            '/pa/%Y/%m/erg_hep_'+level+'_pa_%Y%m%d_'
        if version is None:
            pathformat += 'v??_??.cdf'
        else:
            pathformat += version + '.cdf'

    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True

    if ((level == 'l2') and (datatype == 'omniflux')) or (datatype == '3dflux') or (level == 'l3'):
        # to avoid failure of creation plot variables (at store_data.py) of hep
        notplot = True

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, version=version, force_download=force_download)

    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road

            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ",gatt["PI_AFFILIATION"])
            print('')
            print('- The rules of the road (RoR) common to the ERG project:')
            print(
                '       https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                '- RoR for HEP data: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Hep')
            if level == 'l3':
                print(
                    '- RoR for MGF data: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Mgf')
            print('')
            print('Contact: erg_hep_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    if isinstance(loaded_data, dict):

        if (level == 'l2') and (datatype == 'omniflux'):
            tplot_variables = []
            if prefix + 'FEDO_L' + suffix in loaded_data:
                v_keyname = 'v'
                if v_keyname not in loaded_data[prefix + 'FEDO_L' + suffix]:
                    v_keyname = 'v1'

                v_vars_min = loaded_data[prefix + 'FEDO_L' + suffix][v_keyname][0]
                v_vars_max = loaded_data[prefix + 'FEDO_L' + suffix][v_keyname][1]
                # log average of energy bins
                v_vars = np.power(
                    10., (np.log10(v_vars_min) + np.log10(v_vars_max)) / 2.)
                store_data(prefix + 'FEDO_L' + suffix, data={'x': loaded_data[prefix + 'FEDO_L' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'FEDO_L' + suffix]['y'],
                                                             'v': v_vars},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDO_L' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDO_L' + suffix)

            if prefix + 'FEDO_H' + suffix in loaded_data:
                v_keyname = 'v'
                if v_keyname not in loaded_data[prefix + 'FEDO_H' + suffix]:
                    v_keyname = 'v1'

                v_vars_min = loaded_data[prefix + 'FEDO_H' + suffix][v_keyname][0]
                v_vars_max = loaded_data[prefix + 'FEDO_H' + suffix][v_keyname][1]
                # log average of energy bins
                v_vars = np.power(
                    10., (np.log10(v_vars_min) + np.log10(v_vars_max)) / 2.)
                store_data(prefix + 'FEDO_H' + suffix, data={'x': loaded_data[prefix + 'FEDO_H' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'FEDO_H' + suffix]['y'],
                                                             'v': v_vars},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDO_H' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDO_H' + suffix)

            # remove minus valuse of y array
            if prefix + 'FEDO_L' + suffix in tplot_variables:
                clip(prefix + 'FEDO_L' + suffix, 0., 1.0e+10)
            if prefix + 'FEDO_H' + suffix in tplot_variables:
                clip(prefix + 'FEDO_H' + suffix, 0., 1.0e+10)

            # set spectrogram plot option
            options(prefix + 'FEDO_L' + suffix, 'Spec', 1)
            options(prefix + 'FEDO_H' + suffix, 'Spec', 1)

            # set y axis to logscale
            options(prefix + 'FEDO_L' + suffix, 'ylog', 1)
            options(prefix + 'FEDO_H' + suffix, 'ylog', 1)

            # set yrange
            options(prefix + 'FEDO_L' + suffix, 'yrange', [3.0e+01, 2.0e+03])
            options(prefix + 'FEDO_H' + suffix, 'yrange', [7.0e+01, 2.0e+03])

            # set ytitle
            options(prefix + 'FEDO_L' + suffix, 'ytitle',
                    'HEP-L\nomniflux\nLv2\nEnergy')
            options(prefix + 'FEDO_H' + suffix, 'ytitle',
                    'HEP-H\nomniflux\nLv2\nEnergy')

            # set ysubtitle
            options(prefix + 'FEDO_L' + suffix, 'ysubtitle', '[keV]')
            options(prefix + 'FEDO_H' + suffix, 'ysubtitle', '[keV]')

            # set ylim
            if prefix + 'FEDO_L' + suffix in tplot_variables:
                ylim(prefix + 'FEDO_L' + suffix, 30, 1800)
            if prefix + 'FEDO_H' + suffix in tplot_variables:
                ylim(prefix + 'FEDO_H' + suffix, 500, 2048)

            # set z axis to logscale
            options(prefix + 'FEDO_L' + suffix, 'zlog', 1)
            options(prefix + 'FEDO_H' + suffix, 'zlog', 1)

            # set zrange
            options(prefix + 'FEDO_L' + suffix, 'zrange', [1.0e-15, 1.0e+06])
            options(prefix + 'FEDO_H' + suffix, 'zrange', [1.0e-10, 1.0e+5])

            # set ztitle
            options(prefix + 'FEDO_L' + suffix,
                    'ztitle', '[/cm^{2}-str-s-keV]')
            options(prefix + 'FEDO_H' + suffix,
                    'ztitle', '[/cm^{2}-str-s-keV]')

            # set zlim
            if prefix + 'FEDO_L' + suffix in tplot_variables:
                zlim(prefix + 'FEDO_L' + suffix, 1e+0, 1e+5)
            if prefix + 'FEDO_H' + suffix in tplot_variables:
                zlim(prefix + 'FEDO_H' + suffix, 1e+0, 1e+5)

            # change colormap option
            options(prefix + 'FEDO_L' + suffix,  'Colormap', 'jet')
            options(prefix + 'FEDO_H' + suffix,  'Colormap', 'jet')

            return tplot_variables

        if (level == 'l2') and (datatype == '3dflux'):
            tplot_variables = []
            v2_array = [i for i in range(15)]

            if prefix + 'FEDU_L' + suffix in loaded_data:
                v_keyname = 'v'
                if v_keyname not in loaded_data[prefix + 'FEDU_L' + suffix]:
                    v_keyname = 'v1'
                store_data(prefix + 'FEDU_L' + suffix, data={'x': loaded_data[prefix + 'FEDU_L' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'FEDU_L' + suffix]['y'],
                                                             'v1': np.sqrt(loaded_data[prefix + 'FEDU_L' + suffix][v_keyname][0, :] *
                                                                           loaded_data[prefix + 'FEDU_L' + suffix][v_keyname][1, :]),  # geometric mean for 'v1'
                                                             'v2': v2_array},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_L' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDU_L' + suffix)
                clip(prefix + 'FEDU_L' + suffix, -1.0e+10, 1.0e+10)

            if prefix + 'FEDU_H' + suffix in loaded_data:
                v_keyname = 'v'
                if v_keyname not in loaded_data[prefix + 'FEDU_H' + suffix]:
                    v_keyname = 'v1'


                store_data(prefix + 'FEDU_H' + suffix, data={'x': loaded_data[prefix + 'FEDU_H' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'FEDU_H' + suffix]['y'],
                                                             'v1': np.sqrt(loaded_data[prefix + 'FEDU_H' + suffix][v_keyname][0, :] *
                                                                           loaded_data[prefix + 'FEDU_H' + suffix][v_keyname][1, :]),  # geometric mean for 'v1'
                                                             'v2': v2_array},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_H' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDU_H' + suffix)
                clip(prefix + 'FEDU_H' + suffix, -1.0e+10, 1.0e+10)

            if prefix + 'FEDU_L_Angle_sga' + suffix in loaded_data:
                
                store_data(prefix + 'FEDU_L_Angle_sga' + suffix, data={'x': loaded_data[prefix + 'FEDU_L_Angle_sga' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'FEDU_L_Angle_sga' + suffix]['y']},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_L_Angle_sga' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDU_L_Angle_sga' + suffix)

            if prefix + 'FEDU_H_Angle_sga' + suffix in loaded_data:
                
                store_data(prefix + 'FEDU_H_Angle_sga' + suffix, data={'x': loaded_data[prefix + 'FEDU_H_Angle_sga' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'FEDU_H_Angle_sga' + suffix]['y']},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_H_Angle_sga' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDU_H_Angle_sga' + suffix)

            if prefix + 'FEDU_L_Angle_gse' + suffix in loaded_data:
                
                store_data(prefix + 'FEDU_L_Angle_gse' + suffix, data={'x': loaded_data[prefix + 'FEDU_L_Angle_gse' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'FEDU_L_Angle_gse' + suffix]['y']},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_L_Angle_gse' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDU_L_Angle_gse' + suffix)

            if prefix + 'FEDU_H_Angle_gse' + suffix in loaded_data:
                
                store_data(prefix + 'FEDU_H_Angle_gse' + suffix, data={'x': loaded_data[prefix + 'FEDU_H_Angle_gse' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'FEDU_H_Angle_gse' + suffix]['y']},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_H_Angle_gse' + suffix]['CDF']})
                tplot_variables.append(prefix + 'FEDU_H_Angle_gse' + suffix)

            if prefix + 'sctno_L' + suffix in loaded_data:
                
                store_data(prefix + 'sctno_L' + suffix, data={'x': loaded_data[prefix + 'sctno_L' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'sctno_L' + suffix]['y']},
                           attr_dict={'CDF':loaded_data[prefix + 'sctno_L' + suffix]['CDF']})
                tplot_variables.append(prefix + 'sctno_L' + suffix)

            if prefix + 'sctno_H' + suffix in loaded_data:
                
                store_data(prefix + 'sctno_H' + suffix, data={'x': loaded_data[prefix + 'sctno_H' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'sctno_H' + suffix]['y']},
                           attr_dict={'CDF':loaded_data[prefix + 'sctno_H' + suffix]['CDF']})
                tplot_variables.append(prefix + 'sctno_H' + suffix)

            if prefix + 'rawcnt_H' + suffix in loaded_data:
                v_keyname = 'v'
                if v_keyname not in loaded_data[prefix + 'rawcnt_H' + suffix]:
                    v_keyname = 'v1'

    
                store_data(prefix + 'rawcnt_H' + suffix, data={'x': loaded_data[prefix + 'rawcnt_H' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'rawcnt_H' + suffix]['y'],
                                                             'v1': np.sqrt(loaded_data[prefix + 'rawcnt_H' + suffix][v_keyname][0, :] *
                                                                           loaded_data[prefix + 'rawcnt_H' + suffix][v_keyname][1, :]),  # geometric mean for 'v1'
                                                             'v2': [i for i in range(15)]},
                           attr_dict={'CDF':loaded_data[prefix + 'rawcnt_H' + suffix]['CDF']})
                tplot_variables.append(prefix + 'rawcnt_H' + suffix)

            if prefix + 'rawcnt_L' + suffix in loaded_data:
                v_keyname = 'v'
                if v_keyname not in loaded_data[prefix + 'rawcnt_L' + suffix]:
                    v_keyname = 'v1'

                store_data(prefix + 'rawcnt_L' + suffix, data={'x': loaded_data[prefix + 'rawcnt_L' + suffix]['x'],
                                                             'y': loaded_data[prefix + 'rawcnt_L' + suffix]['y'],
                                                             'v1': np.sqrt(loaded_data[prefix + 'rawcnt_L' + suffix][v_keyname][0, :] *
                                                                           loaded_data[prefix + 'rawcnt_L' + suffix][v_keyname][1, :]),  # geometric mean for 'v1'
                                                             'v2': [i for i in range(15)]},
                           attr_dict={'CDF':loaded_data[prefix + 'rawcnt_L' + suffix]['CDF']})
                tplot_variables.append(prefix + 'rawcnt_L' + suffix)

            return tplot_variables

        if level == 'l3':  # implementation for level = 'l3'

            tplot_variables = []

            if prefix + 'FEDU_L' + suffix in loaded_data:

                L_energy_array_ave = np.sqrt(loaded_data[prefix + 'FEDU_L' + suffix]['v1'][0, :] *
                                             loaded_data[prefix + 'FEDU_L' + suffix]['v1'][1, :])  # geometric mean for 'v1'

                # get energy [keV] array for ytitle options
                L_energy_array = np.trunc(L_energy_array_ave).astype(int)
                non_negative_y_array = np.where(
                    loaded_data[prefix + 'FEDU_L' + suffix]['y'] < 0., np.nan, loaded_data[prefix + 'FEDU_L' + suffix]['y'])
                store_data(prefix + 'FEDU_L' + suffix, data={'x': loaded_data[prefix + 'FEDU_L' + suffix]['x'],
                                                             'y': non_negative_y_array,
                                                             'v1': L_energy_array_ave,
                                                             'v2': loaded_data[prefix + 'FEDU_L' + suffix]['v2']},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_L' + suffix]['CDF']})

                options(prefix + 'FEDU_L' + suffix, 'spec', 1)
                # set ylim
                ylim(prefix + 'FEDU_L' + suffix, 0, 180)
                # set zlim
                zlim(prefix + 'FEDU_L' + suffix, 1e+2, 1e+6)

                tplot_variables.append(prefix + 'FEDU_L' + suffix)

                # make Tplot Variables of erg_hep_l3_FEDU_L_paspec_ene?? (??: 00, 01, 02, ..., 15)
                for i in range(loaded_data[prefix + 'FEDU_L' + suffix]['y'].shape[1]):
                    tplot_name = prefix + 'FEDU_L_paspec_ene' + \
                        str(i).zfill(2) + suffix
                    store_data(tplot_name, data={'x': loaded_data[prefix + 'FEDU_L' + suffix]['x'],
                                                 'y': non_negative_y_array[:, i, :],
                                                 'v': loaded_data[prefix + 'FEDU_L' + suffix]['v2']},
                               attr_dict={'CDF':loaded_data[prefix + 'FEDU_L' + suffix]['CDF']})

                    options(tplot_name, 'spec', 1)
                    # set ylim
                    ylim(tplot_name, 0, 180)
                    # set zlim
                    zlim(tplot_name, 1e+2, 1e+6)
                    # set ytitle
                    options(
                        tplot_name, 'ytitle', f'HEP-L\nEne{str(i).zfill(2)}\n{L_energy_array[i]} keV')

                    tplot_variables.append(tplot_name)

            if prefix + 'FEDU_H' + suffix in loaded_data:

                H_energy_array_ave = np.sqrt(loaded_data[prefix + 'FEDU_H' + suffix]['v1'][0, :] *
                                             loaded_data[prefix + 'FEDU_H' + suffix]['v1'][1, :])  # geometric mean for 'v1'

                # get energy [keV] array for ytitle options
                H_energy_array = np.trunc(H_energy_array_ave).astype(int)
                non_negative_y_array = np.where(
                    loaded_data[prefix + 'FEDU_H' + suffix]['y'] < 0., np.nan, loaded_data[prefix + 'FEDU_H' + suffix]['y'])
                store_data(prefix + 'FEDU_H' + suffix, data={'x': loaded_data[prefix + 'FEDU_H' + suffix]['x'],
                                                             'y': non_negative_y_array,
                                                             'v1': H_energy_array_ave,
                                                             'v2': loaded_data[prefix + 'FEDU_H' + suffix]['v2']},
                           attr_dict={'CDF':loaded_data[prefix + 'FEDU_H' + suffix]['CDF']})

                options(prefix + 'FEDU_H' + suffix, 'spec', 1)
                # set ylim
                ylim(prefix + 'FEDU_H' + suffix, 0, 180)
                # set zlim
                zlim(prefix + 'FEDU_H' + suffix, 1e+1, 1e+4)

                tplot_variables.append(prefix + 'FEDU_H' + suffix)

                # make Tplot Variables of erg_hep_l3_FEDU_H_paspec_ene?? (??: 00, 01, 02, ..., 10)
                for i in range(loaded_data[prefix + 'FEDU_H' + suffix]['y'].shape[1]):
                    tplot_name = prefix + 'FEDU_H_paspec_ene' + \
                        str(i).zfill(2) + suffix
                    store_data(tplot_name, data={'x': loaded_data[prefix + 'FEDU_H' + suffix]['x'],
                                                 'y': non_negative_y_array[:, i, :],
                                                 'v': loaded_data[prefix + 'FEDU_H' + suffix]['v2']},
                               attr_dict={'CDF':loaded_data[prefix + 'FEDU_H' + suffix]['CDF']})

                    options(tplot_name, 'spec', 1)
                    # set ylim
                    ylim(tplot_name, 0, 180)
                    # set zlim
                    zlim(tplot_name, 1e+1, 1e+4)
                    # set ytitle
                    options(
                        tplot_name, 'ytitle', f'HEP-H\nEne{str(i).zfill(2)}\n{H_energy_array[i]} keV')

                    tplot_variables.append(tplot_name)

            # set z axis to logscale
            options(tplot_variables, 'zlog', 1)
            # change colormap option
            options(tplot_variables, 'colormap', 'jet')
            # set ysubtitle
            options(tplot_variables, 'ysubtitle', 'PA [deg]')
            # set ztitle
            options(tplot_variables, 'ztitle', '[/keV/cm^{2}/sr/s]')

            return tplot_variables

    return loaded_data





import logging
import pyspedas


def set_tplot_options(option, value, old_tplot_opt_glob):
    new_tplot_opt_glob = old_tplot_opt_glob

    if option in ['title_text', 'title']:
        new_tplot_opt_glob['title_text'] = value

    elif option == 'title_size':
        new_tplot_opt_glob['title_size'] = value

    elif option == 'var_label':
        new_tplot_opt_glob['var_label'] = value

    elif option == 'x_range':
        new_tplot_opt_glob['x_range'] = value

    elif option == 'data_gap':
        new_tplot_opt_glob['data_gap'] = value

    elif option == 'axis_font_size':
        new_tplot_opt_glob['axis_font_size'] = value

    elif option == 'xmargin':
        new_tplot_opt_glob['xmargin'] = value

    elif option == 'ymargin':
        new_tplot_opt_glob['ymargin'] = value

    elif option == 'style':
        new_tplot_opt_glob['style'] = value

    elif option == 'charsize':
        new_tplot_opt_glob['charsize'] = value

    elif option == 'xsize':
        new_tplot_opt_glob['xsize'] = value

    elif option == 'ysize':
        new_tplot_opt_glob['ysize'] = value

    elif option == 'varlabel_style':
        new_tplot_opt_glob['varlabel_style'] = value

    else:
        logging.warning("Unknown option supplied: " + str(option))

    return new_tplot_opt_glob


def tplot_options(option, value):
    """
    This function allows the user to set GLOBAL options for the generated plots.
    
    Parameters
    ----------
        option : str
            The name of the option.  See section below  
        value : str/int/float/list
            The value of the option.  See section below.  
            
    Options
    -------
        ==================  ==========   =====
        Option (Synonyms)   Value type   Notes
        ==================  ==========   =====
        title_text (title)  str          Title of the entire plot (mpl 'suptitle', above any panel titles)
        title_size          int          Font size of the title
        data_gap            int          Number of seconds with consecutive nan values allowed before no interp should occur
        x_range             [flt, flt]   The min and max x_range (time) to be plotted on all plots
        axis_font_size      int          The font size of the axis ticks.  Default is 10.
        charsize            int          The font size for the legend strings
        style               str          The matplotlib plot style to use
        xsize               flt          The size of the plot window in the X dimension (units of inches)
        ysize               flt          The size of the plot window in the Y dimension (units of inches)
        xmargin             [flt, flt]   The width of the left and right margins of the plot (in inches)
        ymargin             [flt, flt]   The height of the top and bottom margins of the plot (in inches)
        annotations         dict         A dictionary of text, positions, xycoords, and other options to be placed on the plot
        ==================  ==========   =====

    Returns
    -------
        None
    
    Examples
    --------
        >>> # Set the plot title
        >>> import pyspedas
        >>> pyspedas.tplot_options('title', 'SWEA Data for Orbit 1563')
    
    """
    
    option = option.lower()
    
    temp = set_tplot_options(option, value, pyspedas.tplot_tools.tplot_opt_glob)
    pyspedas.tplot_tools.tplot_opt_glob = temp
    
    return




import numpy as np


def mvn_kp_sc_traj_xyz(
    dims_x, dims_y, dims_z, values, x_array, y_array, z_array, nn="linear"
):
    """
    Interpolates the values of a 3D array at given coordinates using trilinear interpolation.

    Parameters:
        dims_x (array-like): The x-coordinates of the grid points.
        dims_y (array-like): The y-coordinates of the grid points.
        dims_z (array-like): The z-coordinates of the grid points.
        values (ndarray): The 3D array of values to be interpolated.
        x_array (array-like): The x-coordinates of the points to interpolate.
        y_array (array-like): The y-coordinates of the points to interpolate.
        z_array (array-like): The z-coordinates of the points to interpolate.
        nn (str, optional): The type of nearest neighbor interpolation to use.
            Defaults to "linear".

    Returns:
        list: The interpolated values at the given coordinates.
    """

    data = []
    if nn == "nearest":
        for x, y, z in np.array([a for a in zip(x_array, y_array, z_array)]):
            ix = np.abs(dims_x - x).argmin()
            iy = np.abs(dims_y - y).argmin()
            iz = np.abs(dims_z - z).argmin()
            data.append(values[ix, iy, iz])
    else:
        max_x = np.max(x_array)
        min_x = np.min(x_array)
        max_y = np.max(y_array)
        min_y = np.min(y_array)
        max_z = np.max(z_array)
        min_z = np.min(z_array)

        for x, y, z in np.array([a for a in zip(x_array, y_array, z_array)]):
            if x > max_x:
                data.append(np.nan)
            elif x < min_x:
                data.append(np.nan)
            elif y > max_y:
                data.append(np.nan)
            elif y < min_y:
                data.append(np.nan)
            elif z > max_z:
                data.append(np.nan)
            elif z < min_z:
                data.append(np.nan)

            sorted_x_distance = np.argsort(np.abs(dims_x - x))
            ix1 = dims_x[sorted_x_distance[0]]
            ix2 = dims_x[sorted_x_distance[1]]
            if ix2 < ix1:
                temp = ix2
                ix2 = ix1
                ix1 = temp
            sorted_y_distance = np.argsort(np.abs(dims_y - y))
            iy1 = dims_y[sorted_y_distance[0]]
            iy2 = dims_y[sorted_y_distance[1]]
            if iy2 < iy1:
                temp = iy2
                iy2 = iy1
                iy1 = temp
            sorted_z_distance = np.argsort(np.abs(dims_z - z))
            iz1 = dims_z[sorted_z_distance[0]]
            iz2 = dims_z[sorted_z_distance[1]]
            if iz2 < iz1:
                temp = iz2
                iz2 = iz1
                iz1 = temp

            nx = (x - ix1) / (ix2 - ix1)
            ny = (y - iy1) / (iy2 - iy1)
            nz = (z - iz1) / (iz2 - iz1)

            data.append(
                values[sorted_x_distance[0], sorted_y_distance[0], sorted_z_distance[0]]
                * (1 - nx)
                * (1 - ny)
                * (1 - nz)
                + values[
                    sorted_x_distance[1], sorted_y_distance[0], sorted_z_distance[0]
                ]
                * nx
                * (1 - ny)
                * (1 - nz)
                + values[
                    sorted_x_distance[0], sorted_y_distance[1], sorted_z_distance[0]
                ]
                * (1 - nx)
                * ny
                * (1 - nz)
                + values[
                    sorted_x_distance[0], sorted_y_distance[0], sorted_z_distance[1]
                ]
                * (1 - nx)
                * (1 - ny)
                * nz
                + values[
                    sorted_x_distance[1], sorted_y_distance[0], sorted_z_distance[1]
                ]
                * nx
                * (1 - ny)
                * nz
                + values[
                    sorted_x_distance[0], sorted_y_distance[1], sorted_z_distance[1]
                ]
                * (1 - nx)
                * ny
                * nz
                + values[
                    sorted_x_distance[1], sorted_y_distance[1], sorted_z_distance[0]
                ]
                * nx
                * ny
                * (1 - nz)
                + values[
                    sorted_x_distance[1], sorted_y_distance[1], sorted_z_distance[1]
                ]
                * nx
                * ny
                * nz
            )
    return data



from .load import load
from typing import List, Union

# This routine was moved out of __init__.py.  Please see that file for previous revision history.

def hiscale(trange:List[str]=['2003-01-01', '2003-01-02'],
        datatype:str='lmde_m1',
        prefix:str='',
        suffix:str='',
        get_support_data:bool=False,
        varformat:str=None,
        varnames:List[str]=[],
        downloadonly:bool=False,
        force_download:bool=False,
        notplot:bool=False,
        no_update:bool=False,
        time_clip:bool=True) -> List[str]:
    """
    This function loads data from the HI-SCALE experiment from the Ulysses mission
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2003-01-01', '2003-01-02']

        datatype: str
            Data type; Valid options: 'lf15_m1', 'lf60_m1', 'lm12_m1', 'lm30_m1', 'lmde_m1', 'wart_m1', 'wrtd_m1'
            Default: 'lmde_m1'

        prefix: str
            The tplot variable names will be given this prefix.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.
            Default: ''

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: '' (all variables loaded)

        varnames: list of str
            List of variable names to load (If empty list or not specified,
            all data variables are loaded)
            Default: []

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        force_download: bool
            Set this flag to download the CDF files, even if local file is newer
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

    Returns
    -------
        list of str
            List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> hiscale_vars = pyspedas.projects.ulysses.hiscale()
    >>> tplot('Electrons')
    """
    return load(instrument='hiscale', trange=trange, datatype=datatype, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, force_download=force_download, notplot=notplot, time_clip=time_clip, no_update=no_update)




import pyspedas
import logging
from pyspedas.tplot_tools import str_to_float_fuzzy, rgb_color

def timebar_delete(t, varname=None, dim='height'):
    if varname is None:
        for name in pyspedas.tplot_tools.data_quants:
            if isinstance(pyspedas.tplot_tools.data_quants[name], dict):
                # non-record varying variable
                continue
            try:
                list_timebars = pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['time_bar']
            except KeyError:
                continue
            elem_to_delete = []
            for elem in list_timebars:
                for num in t:
                    if (elem['location'] == num) and (elem['dimension'] == dim):
                        elem_to_delete.append(elem)
            for i in elem_to_delete:
                list_timebars.remove(i)
            pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['time_bar'] = list_timebars
    else:
        if not isinstance(varname, list):
            varname = [varname]
        for i in varname:
            if i not in pyspedas.tplot_tools.data_quants.keys():
                logging.info(str(i) + " is currently not in pyspedas.")
                return
            if isinstance(pyspedas.tplot_tools.data_quants[i], dict):
                # non-record varying variable
                continue
            try:
                list_timebars = pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['time_bar']
            except KeyError:
                continue
            elem_to_delete = []
            for elem in list_timebars:
                for num in t:
                    if (elem['location'] == num) and (elem['dimension'] == dim):
                        elem_to_delete.append(elem)
            for j in elem_to_delete:
                list_timebars.remove(j)
            # pyspedas.tplot_tools.data_quants[i].attrs['plot_options']['time_bar']
    return


def timebar(t, varname=None, databar=False, delete=False, color="black", thick=1, dash=False):
    """
    This function will add a vertical bar to all time series plots. This is useful if you
    want to bring attention to a specific time.

    Parameters
    ----------
    t : float or list
        The time in seconds since Jan 01 1970 to place the vertical bar. If a list of numbers are supplied,
        multiple bars will be created. If "databar" is set, then "t" becomes the point on the y axis to
        place a horizontal bar.
    varname : str or list, optional
        The variable(s) to add the vertical bar to. If not set, the default is to add it to all current plots. (wildcards accepted)
    databar : bool, optional
        This will turn the timebar into a horizontal data bar. If this is set True, then variable "t" becomes
        the point on the y axis to place a horizontal bar.
    delete : bool, optional
        If set to True, at least one varname must be supplied. The timebar at point "t" for variable "varname"
        will be removed.
    color : str
        The color of the bar.
    thick : int
        The thickness of the bar.
    dash : bool
        If set to True, the bar is dashed rather than solid.

    Returns
    -------
    None

    Examples
    --------
    >>> # Place a green time bar at 2017-07-17 00:00:00
    >>> import pyspedas
    >>> pyspedas.timebar(1500249600, color='green')

    >>> # Place a dashed data bar at 5500 on the y axis
    >>> pyspedas.timebar(5500, dash=True, databar=True)

    >>> # Place 3 magenta time bars of thickness 5
    >>> # at [2015-12-26 05:20:01, 2015-12-26 08:06:40, 2015-12-26 08:53:19]
    >>> # for variable 'sgx' plot
    >>> pyspedas.timebar([1451107201,1451117200,1451119999],'sgx',color='m',thick=5)
    """

    # wildcard expansion
    varname = pyspedas.tnames(varname)

    # make sure t entered is a list
    if not isinstance(t, list):
        t = [t]

    # if entries in list not numerical, run str_to_int
    if not isinstance(t[0], (int, float, complex)):
        t1 = []
        for time in t:
            t1.append(str_to_float_fuzzy(time))
        t = t1

    dim = "height"
    if databar is True:
        dim = "width"

    dash_pattern = "solid"
    if dash is True:
        dash_pattern = "dashed"

    if delete is True:
        timebar_delete(t, varname, dim)
        return

    # if no varname specified, add timebars to every plot
    if varname is None:
        num_bars = len(t)
        for i in range(num_bars):
            tbar = {}
            tbar["location"] = t[i]
            tbar["dimension"] = dim
            tbar["line_color"] = rgb_color(color)[0]
            tbar["line_width"] = thick
            tbar["line_dash"] = dash_pattern
            for name in pyspedas.tplot_tools.data_quants:
                temp_data_quants = pyspedas.tplot_tools.data_quants[name]
                if isinstance(temp_data_quants, dict):
                    # non-record varying variable
                    continue
                temp_data_quants.attrs["plot_options"]["time_bar"].append(tbar)
    # if varname specified
    else:
        if not isinstance(varname, list):
            varname = [varname]
        for j in varname:
            if j not in pyspedas.tplot_tools.data_quants.keys():
                logging.info(str(j) + "is currently not in pyspedas")
            else:
                num_bars = len(t)
                for i in range(num_bars):
                    tbar = {}
                    tbar["location"] = t[i]
                    tbar["dimension"] = dim
                    tbar["line_color"] = rgb_color(color)[0]
                    tbar["line_width"] = thick
                    tbar["line_dash"] = dash_pattern
                    temp_data_quants = pyspedas.tplot_tools.data_quants[j]
                    if isinstance(temp_data_quants, dict):
                        # non-record varying variable
                        continue
                    temp_data_quants.attrs["plot_options"]["time_bar"].append(tbar)
    return



import os
import unittest
from pyspedas.tplot_tools import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_mag_data(self):
        out_vars = pyspedas.projects.st5.mag(time_clip=True)
        self.assertTrue(data_exists('B_SM'))

    def test_load_notplot(self):
        out_vars = pyspedas.projects.st5.mag(notplot=True)
        self.assertTrue('B_SM' in out_vars)

    def test_downloadonly(self):
        files = pyspedas.projects.st5.mag(downloadonly=True, trange=['2006-06-01', '2006-06-02'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()

    


import pandas as pd

# This routine was originally in akebono/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

def load_csv_file(filenames, cols=None, gz=False):
    """
    Loads a list of CSV/txt files into pandas data frames
    """
    if not isinstance(filenames, list):
        filenames = [filenames]
    if gz:
        df = pd.concat((pd.read_csv(f, header=0, sep=r'\s+', dtype=str, names=cols, compression='gzip') for f in filenames), ignore_index=True)
    else:
        df = pd.concat((pd.read_csv(f, header=0, sep=r'\s+', dtype=str, names=cols) for f in filenames), ignore_index=True)
    return df



import numpy as np
from copy import deepcopy
from pyspedas.tplot_tools import get_data, store_data, options, tplot_wildcard_expand
from typing import Union, List
import logging

def _tvectot(tvar: str, new_name: str, join_component: bool):
    data = get_data(tvar)
    md = get_data(tvar,metadata=True)
    new_data = np.linalg.norm(data.y, axis=1)

    if join_component:
        # Caution: this can clobber existing variable names!  Maybe better to manipulate the
        # data directly instead of calling split_vec/join_vec
        #join_vec(split_vec(tvar)+[new_name], newname=new_name)
        joined_data = np.append(data.y, new_data[:,np.newaxis],1)
        store_data(new_name, data={'x':data.times,'y':joined_data}, attr_dict=deepcopy(md))
        options(new_name, 'legend_names', ['x', 'y', 'z', 'Magnitude'])
    else:
        store_data(new_name, data={'x': data.times, 'y': new_data}, attr_dict=md)
        options(new_name, 'legend_names', 'Magnitude')
    return new_name

def tvectot(tvars: Union[str, List[str]],
            newname=None,
            suffix=None,
            join_component=True) -> Union[str , List[str]]:
    """
    Computes the magnitude of a vector time series.

    Parameters
    ----------
    tvars : str or list[str]
        Names of the tplot variables.
    newname: str or list[str]
        Names for the resultant magnitude tplot variables. If not provided, it appends the suffix to `tvars`.
    suffix: str
        The suffix to append to tvars to form newname if newname is not provided.
    join_component: bool
        If True, the magnitude tplot variable is joined with the component tplot variables.

    Returns
    -------
    str or list[str]
        Names of the magnitude tplot variables.
    """

    tvars = tplot_wildcard_expand(tvars)
    if len(tvars) == 0:
        return

    tvars_type = type(tvars)
    if tvars_type == str:
        tvars = [tvars]

    if suffix is None:
        if join_component:
            suffix = "_tot"
        else:
            suffix='_mag'

    if newname is None:
        newname = [tvar + suffix for tvar in tvars]

    # If newname is a scalar string, zip will iterate over characters
    if type(newname) is not list:
        newname = [newname]

    for tvar, newname_single in zip(tvars, newname):
        _tvectot(tvar, newname_single, join_component)
    
    if tvars_type == str:
        return newname[0]
    else:
        return newname



import os
import fnmatch
import glob
import logging
import pandas as pd
from pyspedas.projects.mms.mms_config import CONFIG

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def mms_get_local_ancillary_files(filetype='tetrahedron_qf', trange=None):
    """
    Search for local ancillary files in case a list cannot be retrieved from the
    remote server.  Returns a sorted list of file paths.
    
    Parameters:
        trange : list of str
            time range of interest [start time, end time] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']


        filetype: str
            state file type, e.g. 'tetrahedron_qf' (for tetrahedron quality factor)

    Returns:
        List of local files found matching the input parameters.

    """
    if trange is None:
        logging.info('No trange specified in mms_get_local_ancillary_files')
        return

    # directory and file name search patterns
    # For now
    # -all ancillary data is in one directory:
    #       mms\ancillary
    # -assume file names are of the form:
    #   SPACECRAFT_FILETYPE_startDate_endDate.version
    #   where SPACECRAFT is [MMS1, MMS2, MMS3, MMS4] in uppercase
    #   and FILETYPE is either DEFATT, PREDATT, DEFEPH, PREDEPH in uppercase
    #   and start/endDate is YYYYDOY
    #   and version is Vnn (.V00, .V01, etc..)
    sep = "/" if is_fsspec_uri(CONFIG["local_data_dir"]) else os.path.sep
    dir_pattern = sep.join([CONFIG['local_data_dir'], 'ancillary', 'mms', f'{filetype}'])
    file_pattern = f'MMS_DEFQ_???????_???????.V??'

    files_in_trange = []
    out_files = []

    if is_fsspec_uri(dir_pattern):
        protocol, path = dir_pattern.split("://")
        fs = fsspec.filesystem(protocol)

        files = fs.glob(sep.join([dir_pattern, file_pattern]))
        files = [sep.join([protocol+"://", file]) for file in files]
    else:
        files = glob.glob(sep.join([dir_pattern, file_pattern]))

    for file in files:
        filename = os.path.basename(file) if not is_fsspec_uri(dir_pattern) else file.split("/")[-1]
        try:
            date_parts = filename.split('_')
            start_time_str = date_parts[2]
            end_time_str = date_parts[3].split('.')[0]
            
            start_time = pd.to_datetime(start_time_str, format='%Y%j').timestamp()
            end_time = pd.to_datetime(end_time_str, format='%Y%j').timestamp()
            
            if start_time < pd.Timestamp(trange[1]).timestamp() and end_time >= pd.Timestamp(trange[0]).timestamp():
                files_in_trange.append(file)
        except IndexError:
            continue

    # ensure only the latest version of each file is loaded
    for file in files_in_trange:
        this_file = file[0:-3] + 'V??'
        versions = fnmatch.filter(files_in_trange, this_file)
        if len(versions) > 1:
            out_files.append(sorted(versions)[-1]) # only grab the latest version
        else:
            out_files.append(versions[0])

    return list(set(out_files))



import unittest
from pyspedas.projects.mms.spd_mms_load_bss import spd_mms_load_bss
from pyspedas.projects.mms.mms_load_sroi_segments import mms_load_sroi_segments, get_mms_srois
from pyspedas.projects.mms.deprecated.mms_load_fast_segments import mms_load_fast_segments
from pyspedas.projects.mms.mms_load_brst_segments import mms_load_brst_segments
from pyspedas.projects.mms.mms_update_brst_intervals import mms_update_brst_intervals
from pyspedas.tplot_tools import data_exists, del_data


class SegmentTestCases(unittest.TestCase):
    def test_sroi(self):
        del_data("*")
        sroi = mms_load_sroi_segments(trange=['2019-10-01', '2019-11-01'])
        self.assertTrue(data_exists('mms1_bss_sroi'))
        self.assertTrue(len(sroi[0]) == 28)
        self.assertTrue(sroi[0][0] == 1569849345.0)
        self.assertTrue(sroi[1][0] == 1569923029.0)
        with self.assertLogs(level='ERROR') as log:
            # error, no trange specified
            sroi = mms_load_sroi_segments()
            self.assertIn("Error; no trange specified.", log.output[0])

            # error, start time not specified
            none = get_mms_srois(end_time=1569849345.0)
            self.assertIn("Error, start time not specified", log.output[1])

            # error, end time not specified
            none = get_mms_srois(start_time=1569849345.0)
            self.assertIn("Error, end time not specified", log.output[2])

            # error, probe not specified
            none = get_mms_srois(start_time=1569849345.0, end_time=1569849345.0)
            self.assertIn("Error, sc_id not specified", log.output[3])

    def test_brst(self):
        del_data("*")
        brst = mms_load_brst_segments(trange=['2015-10-16', '2015-10-17'])
        self.assertTrue(len(brst[0]) == 53)
        self.assertTrue(brst[0][0] == 1444975174.0)
        self.assertTrue(brst[1][0] == 1444975244.0)
        self.assertTrue(data_exists('mms_bss_burst'))
        with self.assertLogs(level='ERROR') as log:
            # error, no trange specified
            brst = mms_load_brst_segments()
            self.assertIn("Error; no trange specified.", log.output[0])

    def test_update_brst_intervals(self):
        del_data("*")
        intervals = mms_update_brst_intervals()
        self.assertTrue('start_times' in intervals)
        self.assertTrue('end_times' in intervals)
        self.assertTrue(intervals['start_times'][0] == 1430876725.0)
        self.assertTrue(intervals['end_times'][0] == 1430879715.0)

    def test_spd_mms_load_bss(self):
        del_data("*")
        spd_mms_load_bss(trange=['2015-10-01', '2015-11-01'])
        self.assertTrue(data_exists('mms_bss_burst'))
        spd_mms_load_bss(trange=['2019-10-01', '2019-11-01'])
        self.assertTrue(data_exists('mms_bss_burst'))
        self.assertTrue(data_exists('mms1_bss_sroi'))

    def test_spd_mms_load_bss_err(self):
        del_data("*")
        with self.assertLogs(level='ERROR') as log:
            spd_mms_load_bss(trange=['2015-10-01', '2015-11-01'], datatype='brst', include_labels=True)
            self.assertFalse(data_exists('mms_bss_fast'))
            self.assertIn("Unsupported datatype: brst; valid options: \"fast\" and \"burst\"", log.output[0])


if __name__ == '__main__':
    unittest.main()



import numpy as np
from scipy.integrate import RK45
from geopack import geopack, t89

def t89_trace_to_equator(f, pos_init, max_distance=1e5, dt=1.0):
    """
    Trace the magnetic field line using RK45 until the equator is reached.

    Parameters:
    - f: Function that returns the normalized magnetic field direction.
    - pos_init: Initial position [x, y, z].
    - max_distance: Maximum distance to trace before stopping.
    - dt: Initial time step for RK45.

    Returns:
    - positions: List of positions along the traced field line.
    """

    # Initialize the RK45 integrator
    integrator = RK45(f, 0, pos_init, max_distance, max_step=dt)

    # Store the initial position
    positions = [pos_init]

    # Previous radial component
    prev_radial = np.dot(pos_init, f(0, pos_init))

    while integrator.t < max_distance:
        # Integrate one step
        integrator.step()

        # Get the current position
        pos_current = integrator.y

        # Compute the radial component of B using dot product
        radial_component = np.dot(pos_current, f(0, pos_current))

        # Check if the radial component has changed sign
        if radial_component * prev_radial < 0:
            # Crossed the equator
            break

        # Update the previous radial component
        prev_radial = radial_component

        # Store the current position
        positions.append(pos_current)
        print(pos_current, radial_component)

    return np.array(positions)

def trace_equator_89(time, startpos, iopt=3.0):
    ps = geopack.recalc(time)
    b0_t89=t89.t89(iopt,ps,startpos[0],startpos[1],startpos[2])
    b0_igrf=geopack.igrf_gsm(startpos[0],startpos[1],startpos[2])
    b_init=np.array(b0_t89)+np.array(b0_igrf)
    radial_component=np.dot(b_init,startpos)
    if radial_component < 0.0:
        direction=-1.0 # Field points inward, go the opposite direction
    else:
        direction=1.0 # Field points outward, follow that direction

    def t89_closure(t,pos):
        b_igrf = np.array(geopack.igrf_gsm(pos[0],pos[1],pos[2]))
        b_t89 = np.array(t89.t89(iopt,ps, pos[0],pos[1],pos[2]))
        b=b_igrf+b_t89
        return b*direction/np.linalg.norm(b)

    trace_points = t89_trace_to_equator(t89_closure,pos_init=startpos,max_distance=100.0,dt=0.05)
    equator_point = trace_points[-1]
    print(equator_point)



from .load import load


def emfisis(trange=['2018-11-5', '2018-11-6'],
            probe='a',
            datatype='magnetometer',
            level='l3',
            cadence='4sec',  # for EMFISIS mag data
            coord='sm',  # for EMFISIS mag
            wavetype='waveform',  # for EMFISIS waveform data
            prefix='',
            suffix='',
            force_download=False,
            get_support_data=False,
            varformat=None,
            varnames=[],
            downloadonly=False,
            notplot=False,
            no_update=False,
            time_clip=False):
    """
    This function loads data from the Electric and Magnetic Field Instrument Suite and Integrated Science (EMFISIS) instrument

    For information on the EMFISIS data products, see:
        https://emfisis.physics.uiowa.edu/data/level_descriptions
        https://emfisis.physics.uiowa.edu/data/L2_products


    Parameters
    ----------
        trange : list of str, default=['2018-11-5', '2018-11-6']
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str, default='a'
            Spacecraft probe name: 'a' or 'b'

        datatype : str, default='magnetometer'
            Data type with options varying by data level.
            Level 1:

                'magnetometer'
                'hfr'
                'housekeeping'
                'sc-hk'
                'spaceweather'
                'wfr'
                'wna'

            Level 2:

                'magnetometer'
                'wfr'
                'hfr'
                'housekeeping'

            Level 3:

                'magnetometer'

            Level 4:

                'density'
                'wna-survey'

        level : str, default='l3'
            Data level; options: 'l1', 'l2', 'l3', l4'

        cadence : str, default='4sec'
            Data cadence; options: '1sec', 'hires', '4sec'

        coord : str, default='sm'
            Data coordinate system

        wavetype : str, default='waveform'
            Type of level 2 waveform data with options:

                For WFR data:

                    'waveform' (default)
                    'waveform-continuous-burst'
                    'spectral-matrix'
                    'spectral-matrix-diagonal'
                    'spectral-matrix-diagonal-merged'

                For HFR data:

                    'waveform'
                    'spectra'
                    'spectra-burst'
                    'spectra-merged'

        prefix : str, optional
            The tplot variable names will be given this prefix. By default, no prefix is added.

        suffix : str, optional
            Suffix for tplot variable names. By default, no suffix is added.

        force_download : bool, default=False
            Download file even if local version is more recent than server version.

        get_support_data : bool, default=False
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot. By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat : str, optional
            The file variable formats to load into tplot. Wildcard character
            "*" is accepted. By default, all variables are loaded in.

        varnames: list of str, optional
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool, default=False
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool, default=False
            Return the data in hash tables instead of creating tplot variables

        no_update: bool, default=False
            If set, only load data from your local cache

        time_clip: bool, default=False
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    -------
    tvars : dict or list
        List of created tplot variables or dict of data tables if notplot is True.

    Examples
    --------
    >>> emfisis_vars = pyspedas.projects.rbsp.emfisis(trange=['2018-11-5/10:00', '2018-11-5/15:00'], datatype='magnetometer', level='l3', time_clip=True)
    >>> tplot(['Mag', 'Magnitude'])
    """
    return load(instrument='emfisis', wavetype=wavetype, trange=trange, probe=probe, datatype=datatype, level=level, cadence=cadence, coord=coord, prefix=prefix, suffix=suffix, force_download=force_download, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)




#Import pyspedas tools into pyspedas namespace
# These are imported as a convenience for pyspedas users.  For internal pyspedas development, it is
# probably best to keep using the fully qualified module names.

#from .pytplot import *

from .tplot_tools import get_y_range
from .tplot_tools import tplot_rename
from .tplot_tools import del_data
from .tplot_tools import store_data
from .tplot_tools import store
from .tplot_tools import get_data
from .tplot_tools import get
from .tplot_tools import str_to_float_fuzzy
from .tplot_tools import tplot_names
from .tplot_tools import wildcard_expand
from .tplot_tools import tplot_wildcard_expand
from .tplot_tools import tname_byindex
from .tplot_tools import tindex_byname
from .tplot_tools import options
from .tplot_tools import xlim
from .tplot_tools import ylim
from .tplot_tools import zlim
from .tplot_tools import tlimit
from .tplot_tools import tplot_save
from .tplot_tools import tnames
from .tplot_tools import convert_tplotxarray_to_pandas_dataframe
from .tplot_tools import tplot_restore
from .tplot_tools import is_pseudovariable
from .tplot_tools import count_traces
from .tplot_tools import get_timespan
from .tplot_tools import tplot_options
from .tplot_tools import tplot_rename
from .tplot_tools import tplot_copy
from .tplot_tools import replace_data
from .tplot_tools import replace_metadata
from .tplot_tools import get_ylimits
from .tplot_tools import rgb_color
from .tplot_tools import timebar
from .tplot_tools import databar
from .tplot_tools import del_data
from .tplot_tools import timespan
from .tplot_tools import timestamp
from .tplot_tools import time_float
from .tplot_tools import time_double
from .tplot_tools import time_float_one
from .tplot_tools import time_string_one
from .tplot_tools import time_string
from .tplot_tools import time_datetime
from .tplot_tools import sts_to_tplot
from .tplot_tools import set_coords
from .tplot_tools import get_coords
from .tplot_tools import set_units
from .tplot_tools import get_units
from .tplot_tools import data_exists
from .tplot_tools import link
from .tplot_tools import tres

from .tplot_tools import tinterp
from .tplot_tools import add_across
from .tplot_tools import add
from .tplot_tools import avg_res_data
from .tplot_tools import clip
from .tplot_tools import crop
from .tplot_tools import deflag
from .tplot_tools import degap
from .tplot_tools import derive
from .tplot_tools import divide
from .tplot_tools import interp_nan
from .tplot_tools import join_vec
from .tplot_tools import multiply
from .tplot_tools import spec_mult
from .tplot_tools import split_vec
from .tplot_tools import subtract
from .tplot_tools import pwr_spec
from .tplot_tools import clean_spikes
from .tplot_tools import tdotp
from .tplot_tools import tcrossp
from .tplot_tools import tnormalize
from .tplot_tools import subtract_median
from .tplot_tools import subtract_average
from .tplot_tools import smooth
from .tplot_tools import tsmooth
from .tplot_tools import time_clip
from .tplot_tools import tkm2re
from .tplot_tools import makegap
from .tplot_tools import tdeflag
from .tplot_tools import pwrspc
from .tplot_tools import tpwrspc
from .tplot_tools import dpwrspc
from .tplot_tools import tdpwrspc


from .tplot_tools import reduce_spec_dataset
from .tplot_tools import lineplot
from .tplot_tools import specplot
from .tplot_tools import specplot_make_1d_ybins
from .tplot_tools import get_var_label_ticks
from .tplot_tools import var_label_panel
from .tplot_tools import ctime
from .tplot_tools import highlight
from .tplot_tools import annotate
from .tplot_tools import tplot
from .tplot_tools import cdf_to_tplot
from .tplot_tools import netcdf_to_tplot
from .tplot_tools import tplot_ascii



from .analysis.avg_data import avg_data
from .analysis.deriv_data import deriv_data
from .analysis.tvectot import tvectot
from .analysis.tinterpol import tinterpol
from .analysis.yclip import yclip
from .analysis.twavpol import twavpol
from .analysis.wavelet import wavelet
from .analysis.wavelet98 import wavelet98
from .analysis.wave_signif import wave_signif
from .analysis.wav_data import wav_data
from .analysis.wavelet2 import wavelet2
from .analysis.time_domain_filter import time_domain_filter
from .analysis.find_magnetic_nulls import find_magnetic_nulls_fote, classify_null_type
from .analysis.lingradest import lingradest
from .cdagui_tools.cdagui import cdagui
from .cdagui_tools.cdaweb import CDAWeb
from .cotrans_tools.cotrans import cotrans
from .cotrans_tools.rotmat_get_coords import rotmat_get_coords
from .cotrans_tools.rotmat_set_coords import rotmat_set_coords
from .cotrans_tools.tvector_rotate import tvector_rotate
from .cotrans_tools.cart2spc import cart2spc
from .cotrans_tools.spc2cart import spc2cart
from .cotrans_tools.sphere_to_cart import sphere_to_cart
from .cotrans_tools.cart_to_sphere import cart_to_sphere
from .cotrans_tools.sm2mlt import sm2mlt
from .cotrans_tools.fac_matrix_make import fac_matrix_make
from .cotrans_tools.gsm2lmn import gsm2lmn
from .cotrans_tools.minvar import minvar
from .cotrans_tools.minvar_matrix_make import minvar_matrix_make
from .cotrans_tools.quaternions import qtom, qconj, qdotp, qmult, qnorm, qslerp, qcompose, qvalidate, qdecompose, mtoq, qnormalize
from .cotrans_tools.tvector_rotate import tvector_rotate
from .cotrans_tools.xyz_to_polar import xyz_to_polar
# Importing geopack causes IGRF coefficients to be loaded by the external geopack package, which may not be desired.
#from .geopack.get_tsy_params import get_tsy_params
#from .geopack.get_w_params import get_w
#from .geopack.kp2iopt import kp2iopt
#from .geopack.t01 import t01, tt01
#from .geopack.t89 import t89, tt89
#from .geopack.t96 import t96, tt96
#from .geopack.ts04 import tts04
from .hapi_tools.hapi import hapi

from .particles.moments import moments_3d, spd_pgs_moments, spd_pgs_moments_tplot
from .particles.spd_part_products import spd_pgs_do_fac, spd_pgs_regrid, spd_pgs_v_shift
from .particles.spd_slice2d import slice1d_plot, slice2d, slice2d_plot
from .utilities.spice.time_ephemeris import time_ephemeris
from .utilities.dailynames import dailynames
from .utilities.datasets import find_datasets
# Note: "download" and "download_file" might be problematic names to import, due to risk of conflict with other packages
from .utilities.download import download, download_file, check_downloaded_file
from .utilities.download_ftp import download_ftp
from .utilities.find_ip_address import find_ip_address
from .utilities.interpol import interpol
from .utilities.leap_seconds import load_leap_table
from .utilities.libs import libs
from .utilities.mpause_2 import mpause_2
from .utilities.mpause_t96 import mpause_t96
from .utilities.tcopy import tcopy
from .utilities.is_gzip import is_gzip
from .utilities.xdegap import xdegap
from .version import version


from .projects.noaa.noaa_load_kp import noaa_load_kp
# omni must precede mms to avoid problems with circular imports
from .projects import omni
# lmn_matrix_make requires omni 
from .cotrans_tools.lmn_matrix_make import lmn_matrix_make
from .projects.kompsat.load import load as kompsat_load

# Import routine names with mission prefixes into pyspedas namespace
from .projects.mms import mms_load_mec, mms_load_fgm, mms_load_scm, mms_load_edi, \
    mms_load_edp, mms_load_eis, mms_load_feeps, \
    mms_load_hpca, mms_load_fpi, mms_load_aspoc, \
    mms_load_dsp, mms_load_fsm, mms_load_state, \
    mms_qcotrans, mms_cotrans_lmn, mms_cotrans_qrotate, mms_cotrans_qtransformer
from .projects.mms.feeps_tools.mms_feeps_pad import mms_feeps_pad
from .projects.mms.feeps_tools.mms_feeps_gpd import mms_feeps_gpd
from .projects.mms.eis_tools.mms_eis_pad import mms_eis_pad
from .projects.mms.hpca_tools.mms_hpca_calc_anodes import mms_hpca_calc_anodes
from .projects.mms.hpca_tools.mms_hpca_spin_sum import mms_hpca_spin_sum
from .projects.mms.plots.mms_overview_plot import mms_overview_plot
from .projects.mms.particles.mms_part_getspec import mms_part_getspec
from .projects.mms.particles.mms_part_slice2d import mms_part_slice2d


from . import vires

# set up logging/console output
import logging
from os import environ

logging_level = environ.get('PYSPEDAS_LOGGING_LEVEL')
logging_format = environ.get('PYSPEDAS_LOGGING_FORMAT')
logging_date_fmt = environ.get('PYSPEDAS_LOGGING_DATE_FORMAT')

if logging_format is None:
    logging_format = '%(asctime)s: %(message)s'

if logging_date_fmt is None:
    logging_date_fmt = '%d-%b-%y %H:%M:%S'

if logging_level is None:
    logging_level = logging.INFO
else:
    logging_level = logging_level.lower()
    if logging_level == 'debug':
        logging_level = logging.DEBUG
    elif logging_level == 'info':
        logging_level = logging.INFO
    elif logging_level == 'warn' or logging_level == 'warning':
        logging_level = logging.WARNING
    elif logging_level == 'error':
        logging_level = logging.ERROR
    elif logging_level == 'critical':
        logging_level = logging.CRITICAL

logging.captureWarnings(True)

# basicConfig here doesn't work if it has previously been called
logging.basicConfig(format=logging_format, datefmt=logging_date_fmt, level=logging_level)

# manually set the logger options from the defaults/environment variables
logger = logging.getLogger()
logger_handler = logger.handlers[0]  # should exist since basicConfig has been called
logger_fmt = logging.Formatter(logging_format, logging_date_fmt)
logger_handler.setFormatter(logger_fmt)
logger.setLevel(logging_level)

# Now that logging is configured, check to see if pytplot is installed.  If so, advise the
# user that pytplot is now included with pyspedas, and that the external pytplot package can
# and should be removed.

try:
    import tplot_tools as old_pytplot
    # If we get here, the external package is installed.
    logging.info("You appear to have the external pytplot or pytplot-mpl-temp packages installed.")
    logging.info("PySPEDAS 2.0 and later now bundles the pytplot tools, and we recommend that you remove the external packages to avoid using obsolete code.")
    logging.info("To uninstall, run these commands in a terminal window (not a Python window!) in the environment you're using with PySPEDAS:")
    logging.info("pip uninstall pytplot")
    logging.info("pip uninstall pytplot-mpl-temp")
    logging.info("To update your own code for PySPEDAS 2.0 compatibility, please refer to our PySPEDAS 2.0 migration guide:")
    logging.info("https://pyspedas.readthedocs.io/en/latest/pyspedas_2_migration.html")
except ImportError:
    # Nothing to do here, external pytplot package not found
    pass



import os
import re
import sys
import warnings
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import logging
import fnmatch
import datetime
import fsspec
from importlib.metadata import version, PackageNotFoundError

from pathlib import Path
from shutil import copyfileobj, copy
from tempfile import NamedTemporaryFile
from html.parser import HTMLParser
from netCDF4 import Dataset
from cdflib import CDF
from time import sleep
from .rate_connection_quality import rate_connection_quality

def is_fsspec_uri(uri):
    """
    See if uri is something fsspec can handle.
    Pulled from Pandas:
        https://github.com/pandas-dev/pandas/blob/main/pandas/io/common.py#L291
    """
    _RFC3896_ = re.compile(r"^[A-Za-z][A-Za-z0-9+\-+.]*://")
    return (
            isinstance(uri, str)
            and bool(_RFC3896_.match(uri))
            and not uri.startswith(("http://", "https://"))
    )

class LinkParser(HTMLParser):
    """
    A custom HTML parser to extract links from an HTML document.

    This class is a subclass of HTMLParser from the html.parser module. It overrides the handle_starttag method to
    extract 'href' attributes from 'a' tags, which represent links in an HTML document.

    Attributes
    ----------
    links : list
        List of links extracted from the HTML document.

    Methods
    -------
    handle_starttag(tag, attrs)
        Handle the start of an HTML tag.
    """

    def handle_starttag(self, tag, attrs):
        """
        Handle the start of an HTML tag.

        If the tag is an 'a' tag, this method extracts the 'href' attribute (if present) and adds it to the links list.

        Parameters
        ----------
        tag : str
            Name of the HTML tag.
        attrs : list of (str, str) tuples
            List of (name, value) pairs containing the attributes of the HTML tag.

        Notes
        -----
        This method is called by the HTMLParser feed method for each start tag encountered in the HTML document.
        """
        if tag == "a":
            attrs = {k: v for (k, v) in attrs}
            if "href" in attrs:
                link = attrs["href"]
                # kludge to support http://rbspice?.ftecs.com/
                if "/" in link:
                    link = link.split("/")[-1]
                try:
                    self.links.append((link))
                except AttributeError:
                    self.links = [(link)]


def check_downloaded_file(filename):
    """
    Check if a file exists and if it can be opened (for CDF and netCDF files).

    If the file exists but it is not CDF or netCDF, it returns True without trying to open the file.

    Parameters
    ----------
    filename : str
        Name of the file to check.

    Returns
    -------
    bool
        True if the file exists and can be opened, False otherwise.

    Notes
    -----
    This function specifically checks for CDF and netCDF files. If the file is of a different type,
    it simply checks for its existence without trying to open it.
    """
    result = False

    if is_fsspec_uri(filename):
        # cdflib only supports AWS URI currently
        # https://github.com/MAVENSDC/cdflib/blob/main/cdflib/cdfread.py#L77
        try:
            cdf_file = CDF(filename)
            return True
        except (ImportError, ModuleNotFoundError) as ie:
            logging.error("ERROR: " + str(ie))
            sys.exit(1)
        except Exception as e:
            logging.info("Cannot open CDF file: " + filename)
            logging.error(e)
            return result

    fpath = Path(filename)
    if fpath.is_file() and len(filename) > 3:
        if filename[-4:] == ".cdf":
            # Try to open the cdf file
            try:
                cdf_file = CDF(filename)
                result = True
            except:
                logging.info("Cannot open CDF file: " + filename)
                result = False
        elif filename[-3:] == ".nc":
            # Try to open the netCDF file
            try:
                netcdf_file = Dataset(filename)
                result = True
            except:
                logging.info("Cannot open netCDF file: " + filename)
                result = False
        else:
            # The file is not CDF or netCDF, issue a debug-level log message and return true
            logging.debug("The file is not CDF or netCDF. Filename: " + filename)
            result = True

    return result


def download_file(
    url=None,
    filename=None,
    headers={},
    username=None,
    password=None,
    verify=False,
    session=None,
    basic_auth=False,
    nbr_tries=0,
    text_only=None,
    force_download=False
):
    """
    Download a file and return its local path; this function is primarily meant to be called by the download function.

    Parameters
    ----------
    url : str
        Remote URL to download.
    filename : str
        Local file name.
    headers : dict
        Dictionary containing the headers to be passed to the requests get call.
    username : str, optional
        Username to be used in HTTP authentication.
    password : str, optional
        Password to be used in HTTP authentication.
    verify : bool, optional
        Flag indicating whether to verify the SSL/TLS certificate.
    session : requests.Session object, optional
        Requests session object that allows you to persist things like HTTP authentication through multiple calls.
    basic_auth : bool, optional
        Flag to indicate that the remote server uses basic authentication instead of digest authentication.
    nbr_tries : int, optional
        Counts how many times we tried to download the file. Default is 0.
    text_only : bool, optional
        If True, save the response content as utf-8 encoded text.  If False, treat as a stream of bytes. If None,
        try to infer the correct format from the filename suffix. Defaults to None.
    force_download : bool, optional
        Flag to indicate if the file should be downloaded even if a local version exists.
        This causes the local version of the file to be overwritten.

    Returns
    -------
    str
        String containing the local file name.

    Notes
    -----
    Checks if the CDF or netCDF file can be opened, and if it can't, tries to download the file for a second time.

    In support of URIs passed as inputs, priority is given to streaming from a URI source
    rather than downloading the file(s). This behavior is separate from web-based urls
    that would become overloaded. The package fsspec is used to simulate a filesystem
    functionality for several URI protocols.
    """
    headers_original = headers
    session_original = session

    # Cloud Awareness
    if is_fsspec_uri(url):
        protocol, path = url.split("://")
        remote_fs = fsspec.filesystem(protocol, anon=False)
        remote_modtime = remote_fs.info(path)["LastModified"]

        if is_fsspec_uri(filename): # two URIs given
            local_protocol, local_path = filename.split("://")
            local_fs = fsspec.filesystem(local_protocol, anon=False)

            if local_fs.exists(filename):
                local_modtime = local_fs.info(local_path)["LastModified"]

                # local newer than remote
                if local_modtime >= remote_modtime:
                    logging.info("Streaming from local URI (current): " + filename)
                    return filename

            # priority to stream from remote URI
            if not force_download:
                logging.info("Streaming from remote: " + url)
                return url

            # save remote file to host os and push to uri referred to as "local"
            logging.info("Retrieving newer remote file: " + url)
            remote_fs.get(url, os.getcwd())
            local_path = os.path.join(os.getcwd(), url[url.rfind("/")+1:])
            local_fs.put(local_path, filename)
            logging.info("File placed on \"local\": " + filename)
            os.remove(local_path) # cleanup on host os

            return filename

        # filename is not URI
        if os.path.exists(filename):
            if force_download:
                logging.info("Retrieving remote file: " + url)
                remote_fs.get(url, filename)
                logging.info("File placed on \"local\": " + filename)
                return filename
            local_modtime = datetime.datetime.fromtimestamp(os.path.getmtime(filename), datetime.timezone.utc)

            # remote newer than local
            if local_modtime < remote_modtime:
                logging.info("Streaming from remote: " + url)
                return url

            logging.info("File is current " + filename)
            return filename

        if not force_download:
            logging.info("Streaming from remote: " + url)
            return url

        # download from remote
        logging.info("Retrieving remote file: " + url)
        remote_fs.get(url, filename)
        logging.info("File placed on \"local\": " + filename)

        return filename

    # from this point, remote is NOT fsspec URI

    # update header from URI
    if is_fsspec_uri(filename):
        protocol, path = filename.split("://")
        fs = fsspec.filesystem(protocol, anon=False)

        if fs.exists(filename) and not force_download:
            mod_tm = (fs.info(path)["LastModified"]).strftime("%a, %d %b %Y %H:%M:%S GMT")
            headers["If-Modified-Since"] = mod_tm
    else:
        # check if the file exists, and if so, set the last modification time in the header
        # this allows you to avoid re-downloading files that haven't changed
        if os.path.exists(filename) and not force_download:
            mod_tm = (
                datetime.datetime.fromtimestamp(
                    os.path.getmtime(filename), datetime.timezone.utc
                )
            ).strftime("%a, %d %b %Y %H:%M:%S GMT")
            headers["If-Modified-Since"] = mod_tm

    if session is None:
        session = requests.Session()
        # Configure retry strategy
        # We'll just use a fixed configuration, unless it turns out to need fine-tuning
        retries = Retry(
            total=3,  # Total number of retries
            backoff_factor=2,  # Exponential backoff factor (sleep for 0s, 4s, 8s between retries)
            status_forcelist=[429, 500, 502, 503, 504],  # HTTP status codes to force a retry on
            allowed_methods=["GET"] # HTTP methods to retry on
        )
        session.mount('http://', HTTPAdapter(max_retries=retries))
        session.mount('https://', HTTPAdapter(max_retries=retries))



    if username is not None:
        session.auth = requests.auth.HTTPDigestAuth(username, password)

    connect_timeout_secs = 10
    read_timeout_secs = 20
    request_time = datetime.datetime.now()
    transfer_bytes = 0
    try:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ResourceWarning)
            if not basic_auth:
                fsrc = session.get(url,
                                   stream=True,
                                   verify=verify,
                                   headers=headers,
                                   timeout=(connect_timeout_secs, read_timeout_secs),
                                   )
            else:
                fsrc = session.get(
                    url,
                    stream=True,
                    verify=verify,
                    headers=headers,
                    auth=(username, password),
                    timeout=(connect_timeout_secs, read_timeout_secs),
                )
    except requests.exceptions.ConnectionError as e:
        logging.error("Unable to connect to URL: %s", str(e))
        return None

    # need to delete the If-Modified-Since header so it's not set in the dictionary in subsequent calls
    if headers.get("If-Modified-Since") is not None:
        del headers["If-Modified-Since"]

    needs_to_download_file = False
    if fsrc.status_code == 304 and not force_download:
        # the file hasn't changed
        logging.info("File is current: " + filename)
        fsrc.close()
    elif fsrc.status_code == 404:
        # file not found
        logging.error("Remote file not found: " + url)
        fsrc.close()
        return None
    elif fsrc.status_code == 401 or fsrc.status_code == 403:
        # authentication issues
        logging.error("Unauthorized: " + url)
        fsrc.close()
        return None
    elif fsrc.status_code == 200 or (fsrc.status_code == 304 and force_download):
        # this is the main download case
        needs_to_download_file = True
        logging.info("Downloading " + url + " to " + filename)
    else:
        # all other problems
        logging.error(fsrc.reason)
        fsrc.close()
        return None

    content_saved_ok = True
    if needs_to_download_file:
        fsuffix = filename.split('.') # could be fsspec uri
        fsuffix = '' if fsuffix[0] == filename else fsuffix[-1]
            
        ftmp = NamedTemporaryFile(delete=False, suffix=fsuffix)

        # If no text_only value was passed, try to determine whether the file is text (and should be utf-8 encoded),
        # or binary (saved as-is) by looking at the file extension.

        if text_only is None:
            if fsuffix.lower() in ['', 'txt', 'csv', 'html', 'tab', 'log', 'rtf', 'md', 'xml', 'pdf']:
                text_only = True
            else:
                text_only = False

        with open(ftmp.name, "wb") as f:
            try:
                transfer_bytes = len(fsrc.content)
                # There is also the fsrc.raw file object, which returns the raw bytes as read from the socket.
                # It may be gzip-compressed.  This is probably the wrong choice in nearly every scenario, so that
                # option has been removed.
                if text_only:
                    ret = f.write(fsrc.text.encode("utf-8"))
                else:
                    ret = f.write(fsrc.content)
            except requests.exceptions.ChunkedEncodingError:
                logging.warning("A ChunkedEncodingError was encountered while saving the request data.  The file may be corrupted.")
                content_saved_ok = False

        completion_time = datetime.datetime.now()
        elapsed_dt = completion_time - request_time
        elapsed_secs = elapsed_dt.total_seconds()
        transfer_mbytes = transfer_bytes / 1024 / 1024
        transfer_rate = transfer_mbytes/elapsed_secs
        # We may want to have more categories here, and explicitly log very slow transfers
        transfer_quality = rate_connection_quality(elapsed_secs, transfer_mbytes, transfer_rate)

        if is_fsspec_uri(filename):
            protocol, path = filename.split("://")
            fs = fsspec.filesystem(protocol, anon=False)
            
            # copy method is within filesystems under fsspec
            if check_downloaded_file(ftmp.name):
                fs.put(ftmp.name, filename)
                logging.info(f"Download of {filename} complete, {transfer_mbytes:.3f} MB in {elapsed_secs:.1f} sec ({transfer_rate:.3f} MB/sec) ({transfer_quality})")
            else:
                logging.error(f"Download of {filename} failed, {transfer_mbytes:.3f} MB in {elapsed_secs:.1f} sec ({transfer_rate:.3f} MB/sec) ({transfer_quality}). The temp file will be removed.")
                logging.error("If the same file has been already downloaded previously, it might be possible to use that instead.")
        else:
            # make sure the directory exists
            if (
                not os.path.exists(os.path.dirname(filename))
                and os.path.dirname(filename) != ""
            ):
                os.makedirs(os.path.dirname(filename))
    
            # if the download was successful, copy to data directory
            if check_downloaded_file(ftmp.name):
                copy(ftmp.name, filename)
                logging.info(f"Download of {filename} complete, {transfer_mbytes:.3f} MB in {elapsed_secs:.1f} sec ({transfer_rate:.3f} MB/sec) ({transfer_quality})")
            else:
                logging.error(f"Download of {filename} failed, {transfer_mbytes:.3f} MB in {elapsed_secs:.1f} sec ({transfer_rate:.3f} MB/sec) ({transfer_quality}). The temp file will be removed.")
                logging.error("If the same file has been already downloaded previously, it might be possible to use that instead.")

        # cleanup
        fsrc.close()
        ftmp.close()
        os.unlink(ftmp.name)  # delete the temporary file
            

    # At this point, we check if the file can be opened.
    # If it cannot be opened, we delete the file and try again.

    # We haven't tried the download a second time yet, so we'll set the second_try flag to non-null
    second_try = False
    if nbr_tries == 0 and (content_saved_ok is False or check_downloaded_file(filename) is False):
        nbr_tries = 1
        logging.info("There was a problem with the file: " + filename)
        logging.info("We are going to download it for a second time.")
        if is_fsspec_uri(filename):
            fs = fsspec.filesystem(protocol, anon=False)
            fs.delete(filename)
        elif os.path.exists(filename):
            os.unlink(filename)

        second_try = download_file(
            url=url,
            filename=filename,
            headers=headers_original,
            username=username,
            password=password,
            verify=verify,
            session=session_original,
            basic_auth=basic_auth,
            nbr_tries=nbr_tries,
            text_only=text_only,
        )

    # If the file again cannot be opened, we give up.
    # Since we did the second download with another call to download_file(), the content_saved_ok flag
    # is no longer valid.  If the recursive call also failed, it should have returned None, so we'll
    # check for that.
    # TODO: Maybe this block should be indented another level, and only run if we made a second attempt?
    if nbr_tries > 0 and (second_try is None or check_downloaded_file(filename) is False):
        nbr_tries = 2
        logging.info("Tried twice. There was a problem with the file: " + filename)
        logging.info("File will be removed. Try to download it again at a later time.")
        if is_fsspec_uri(filename):
            fs = fsspec.filesystem(protocol, anon=False)
            fs.delete(filename)
        elif os.path.exists(filename):
            os.unlink(filename)
        filename = None

    return filename


def download(
    remote_path="",
    remote_file="",
    local_path="",
    local_file="",
    headers={},
    username=None,
    password=None,
    verify=True,
    session=None,
    no_download=False,
    last_version=False,
    basic_auth=False,
    regex=False,
    no_wildcards=False,
    text_only=None,
    force_download=False,
):
    """
    Download one or more remote files and return their local paths.

    Parameters
    ----------
    remote_path : str
        String consisting of a common URL base for all remote files.
    remote_file : str or list of str
        String or string array of URLs to remote files.
    local_path : str
        String consisting of a common local path for all local files.
    local_file : str or list of str
        String or string array of local destination file names.
    headers : dict
        Dictionary containing the headers to be passed to the requests get call.
    username : str, optional
        Username to be used in HTTP authentication.
    password : str, optional
        Password to be used in HTTP authentication.
    basic_auth : bool, optional
        Flag to indicate that the remote server uses basic authentication instead of digest authentication.
    verify : bool, optional
        Flag indicating whether to verify the SSL/TLS certificate.
    session : requests.Session object, optional
        Requests session object that allows you to persist things like HTTP authentication through multiple calls.
    no_download : bool, optional
        Flag to not download remote files.
    last_version : bool, optional (default is True)
        Flag to only download the last in file in a lexically sorted list when multiple matches are found using wildcards.
    regex : bool, optional
        Flag to allow regular expressions in the file name matching, instead of unix style matching.
    no_wildcards : bool, optional
        Flag to assume no wild cards in the requested url/filename.
    text_only : bool, optional
        If True, save the response content as utf-8 encoded text.  If False, treat as a stream of bytes. If None,
        try to infer the correct format from the filename suffix.  Defaults to None.
    force_download : bool, optional
        Flag to indicate if the file should be downloaded even if a local version exists.
        This causes the local version of the file to be overwritten.

    Cloud Awareness
    ---------------
    In support of URIs passed as inputs, priority is given to streaming from a URI source
    rather than downloading the file(s). This behavior is separate from web-based urls
    that would become overloaded. The package fsspec is used to simulate a filesystem
    functionality for several URI protocols.

    Returns
    -------
    list of str
        String list specifying the full local path to all requested files.

    Examples
    --------
    >>> from pyspedas import download
    >>> remote_path = "https://spdf.gsfc.nasa.gov/pub/data/omni/omni_cdaweb/hro_5min/2012/"
    >>> remote_files = ["omni_hro_5min_20121101_v01.cdf", "omni_hro_5min_20121201_v01.cdf"]
    >>> local_path = "/tmp/omni/"
    >>> files = download(remote_path=remote_path, remote_file=remote_files, local_path=local_path)
    >>> print(files)
    ['/tmp/omni/omni_hro_5min_20121101_v01.cdf', '/tmp/omni/omni_hro_5min_20121201_v01.cdf']
    """
    local_file_in = local_file

    if isinstance(remote_path, list):
        logging.error("Remote path must be a string")
        return

    if isinstance(local_path, list):
        logging.error("Local path must be a string")
        return

    if local_path == "":
        local_path = str(Path("").resolve())

    if username is not None and password is None:
        logging.error("Username provided without password")
        return

    if session is None:
        session = requests.Session()
        # Configure retry strategy
        # We'll just use a fixed configuration, unless it turns out to need fine-tuning
        retries = Retry(
            total=3,  # Total number of retries
            backoff_factor=2,  # Exponential backoff factor (sleep for 0s, 4s, 8s between retries)
            status_forcelist=[429, 500, 502, 503, 504],  # HTTP status codes to force a retry on
            allowed_methods=["GET"] # HTTP methods to retry on
        )
        session.mount('http://', HTTPAdapter(max_retries=retries))
        session.mount('https://', HTTPAdapter(max_retries=retries))

    if username is not None:
        session.auth = requests.auth.HTTPDigestAuth(username, password)

    if headers.get("User-Agent") is None:
        try:
            release_version = version("pyspedas")
        except PackageNotFoundError:
            release_version = "bleeding edge"
        headers["User-Agent"] = "pySPEDAS " + release_version

    out = []
    index_table = {}

    # To avoid hammering the remote server with repeated failing requests, if we have a problem with an index
    # URL we'll add it to bad_index_set and skip it if it comes up again.
    bad_index_set = set()

    if not isinstance(remote_file, list):
        remote_file = [remote_file]

    urls = [remote_path + rfile for rfile in remote_file]

    for url in urls:
        resp_data = None
        url_file = url[url.rfind("/") + 1 :]
        url_base = url.replace(url_file, "")

        # automatically use remote_file locally if local_file is not specified
        if local_file_in == "":
            # if remote_file is the entire url then only use the filename
            if remote_path == "":
                local_file = url_file
            else:
                local_file = url.replace(remote_path, "")

                if local_file == "":  # remote_path was the full file name
                    local_file = remote_path[remote_path.rfind("/") + 1 :]

        filename = os.path.join(local_path, local_file)

        short_path = local_file[: 1 + local_file.rfind("/")]

        if not no_download:
            # expand the wildcards in the url
            if ("?" in url or "*" in url or regex) and (
                not no_download and not no_wildcards
            ):
                if index_table.get(url_base) is not None and not is_fsspec_uri(url):
                    links = index_table[url_base]
                elif url_base in bad_index_set:
                    logging.info(
                        "Skipping remote index: "
                        + url_base
                        + " (previous attempt failed)"
                    )
                    continue
                elif is_fsspec_uri(url):
                    # when remote is URI, do not download data / read in place
                    protocol, path = url.split("://")
                    fs = fsspec.filesystem(protocol, anon=False)

                    if not is_fsspec_uri(local_path):
                        if force_download:
                            # obtain the file names to be used in the new remote_file argument
                            # URIs are not Paths so cannot use Path.name or os.path.basename
                            links = [link[link.rfind("/")+1:] for link in fs.glob(url)]
                            index_table[url_base] = links
                        else:
                            links = [protocol + "://" + link for link in fs.glob(url)]

                            if len(links) > 0:
                                for link in links:
                                    logging.info("Using remote URI file: "+link)
                                    out.append(link)
                                continue
                    else:
                        # local is URI so we are just updating files between URIs
                        if index_table.get(url_base) is None:
                            logging.info("Retrieving listings from directory: " + url_base)
                        else:
                            # reset since we glob for specific files instead of a full directory listing
                            index_table = {}
                        links = [link[link.rfind("/")+1:] for link in fs.glob(url)]
                        index_table[url_base] = links
                else:
                    logging.info("Downloading remote index: " + url_base)

                    # we'll need to parse the HTML index file for the file list
                    index_start_time = datetime.datetime.now()
                    connect_timeout_secs = 10
                    read_timeout_secs = 20
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=ResourceWarning)
                        try:
                            if not basic_auth:
                                html_index = session.get(
                                    url_base,
                                    verify=verify,
                                    headers=headers,
                                    timeout=(connect_timeout_secs, read_timeout_secs),
                                )
                            else:
                                html_index = session.get(
                                    url_base,
                                    verify=verify,
                                    headers=headers,
                                    auth=(username, password),
                                    timeout=(connect_timeout_secs, read_timeout_secs),
                                )
                        except requests.exceptions.ConnectionError:
                            # Add this index to bad_index_set and cool down a bit
                            logging.warning(f"Connection error getting remote index {url_base}, marking this URL as bad")
                            bad_index_set.add(url_base)
                            sleep(2)
                            continue

                    if html_index.status_code == 404:
                        logging.error("Remote index not found: " + url_base)
                        # Add this index to bad_index_set and cool down a bit
                        bad_index_set.add(url_base)
                        sleep(2)
                        continue

                    if html_index.status_code == 401 or html_index.status_code == 403:
                        logging.error("Unauthorized: " + url_base)
                        # Add this index to bad_index_set and cool down a bit
                        bad_index_set.add(url_base)
                        sleep(2)
                        continue

                    # grab the links
                    link_parser = LinkParser()
                    link_parser.feed(html_index.text)

                    try:
                        links = link_parser.links
                        index_table[url_base] = links
                    except AttributeError:
                        links = []

                    index_done_time = datetime.datetime.now()
                    index_dt = index_done_time - index_start_time
                    index_elapsed = index_dt.total_seconds()

                    if index_elapsed > 5.0:
                        logging.warning(f"Remote index took {index_elapsed:.1f} seconds to return, may indicate problems on remote server (index_slow)")
                # find the file names that match our string
                if not regex:
                    # note: fnmatch.filter accepts ? (single character) and * (multiple characters)
                    new_links = fnmatch.filter(links, url_file)
                else:
                    reg_expression = re.compile(url_file)
                    new_links = list(filter(reg_expression.match, links))

                if len(new_links) == 0:
                    logging.info(
                        "No links matching pattern %s found at remote index %s",
                        url_file,
                        url_base,
                    )

                if last_version and len(new_links) > 1:
                    new_links = sorted(new_links)
                    new_links = [new_links[-1]]

                if "?" in remote_path or "*" in remote_path:
                    # the user specified a wild card in the remote_path
                    remote_path = url_base

                # download the files
                # Add a small delay between requesting the index and requesting the files
                sleep(1)

                for new_link in new_links:
                    resp_data = download(
                        remote_path=remote_path,
                        remote_file=short_path + new_link,
                        local_path=local_path,
                        username=username,
                        password=password,
                        verify=verify,
                        headers=headers,
                        session=session,
                        basic_auth=basic_auth,
                        text_only=text_only,
                        force_download=force_download
                    )
                    if resp_data is not None:
                        for file in resp_data:
                            out.append(file)
                session.close()
                continue
            resp_data = download_file(
                url=url,
                filename=filename,
                username=username,
                password=password,
                verify=verify,
                headers=headers,
                session=session,
                basic_auth=basic_auth,
                text_only=text_only,
                force_download=force_download
            )

        if resp_data is not None:
            if not isinstance(resp_data, list):
                resp_data = [resp_data]
            for file in resp_data:
                out.append(file)
        else:
            # download wasn't successful, search for local files
            # logging.info("Searching for local files...")

            temp_out = []

            if local_path == "":
                local_path_to_search = str(Path(".").resolve())
            else:
                local_path_to_search = local_path

            local = local_file[local_file.rfind("/") + 1 :]

            # find matching files from URI
            if is_fsspec_uri(local_path_to_search):
                protocol, path = local_path_to_search.split("://")
                fs = fsspec.filesystem(protocol, anon=False)
                walk = fs.walk(local_path_to_search)
            else:
                walk = os.walk(local_path_to_search)

            for dirpath, dirnames, filenames in walk:
                if not regex:
                    matching_files = fnmatch.filter(filenames, local)
                else:
                    reg_expression = re.compile(local)
                    matching_files = list(filter(reg_expression.match, filenames))

                for file in matching_files:
                    if is_fsspec_uri(local_path_to_search):
                        temp_out.append(protocol + "://" + dirpath + '/' + file)
                        logging.info("Streaming from local URI: " + temp_out[-1])
                    else:
                        # out.append(os.path.join(dirpath, file))
                        temp_out.append(os.path.join(dirpath, file))

            # check if the file exists, and if so, set the last modification time in the header
            if len(temp_out) == 0:
                logging.info("No local files found for " + url)
                continue
            temp_out = sorted(temp_out)

            if last_version:
                out.append(temp_out[-1])  # append the latest version
                logging.info(f"Local file found: {temp_out[-1]}")
            else:
                for file in temp_out:
                    out.append(file)
                logging.info(f"Local files found: {temp_out}")

    session.close()
    return out




from .t89 import tt89
from .t96 import tt96
from .t01 import tt01
from .ts04 import tts04
from .get_tsy_params import get_tsy_params
from .kp2iopt import kp2iopt
from .t89_trace_equator import trace_equator_89
from .t89_trace_iono import trace_iono_89



from pyspedas.tplot_tools import options


def mms_scm_set_metadata(tnames, probe, datatype, coord, suffix=''):
    """
    This function updates the metadata for the SCM data products
    
    Parameters
    ----------
        tnames : list of str
            list of tplot variables loaded by the load routine

        probe : str 
            spacecraft probe #, valid values for MMS probes are ['1', '2', '3', '4']. 

        datatype : str 
            SCM datatype loaded

        coord : str or list of str
            SCM coordinate system; depends on data level
            
        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

    """

    if not isinstance(tnames, list): tnames = [tnames]

    for tvar in tnames:
        if tvar == 'mms'+probe+'_scm_'+datatype+'_'+coord+suffix:
            options(tvar, 'ytitle', 'MMS'+probe+' '+datatype+' ('+coord+')')
            options(tvar, 'legend_names', ['1', '2', '3'])
        elif tvar == 'mms'+probe+'_scm_acb_'+coord+'_scsrvy_srvy_l2'+suffix:
            options(tvar, 'legend_names', ['Bx', 'By', 'Bz'])
            options(tvar, 'ytitle', 'MMS'+probe+' SCM '+datatype)
        elif tvar == 'mms'+probe+'_scm_acb_'+coord+'_scb_brst_l2'+suffix:
            options(tvar, 'legend_names', ['Bx', 'By', 'Bz'])
            options(tvar, 'ytitle', 'MMS'+probe+' SCM '+datatype)
        elif tvar == 'mms'+probe+'_scm_acb_'+coord+'_schb_brst_l2'+suffix:
            options(tvar, 'legend_names', ['Bx', 'By', 'Bz'])
            options(tvar, 'ytitle', 'MMS'+probe+' SCM '+datatype)



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/Pytplot

import pyspedas
from pyspedas.tplot_tools import store_data
import numpy as np
import copy
import logging

def derive(tvar,newname=None):
    """
    Takes the derivative of the tplot variable.

    Parameters
    ----------
        tvar : str
            Name of tplot variable.
        newname : str
            Name of new tplot variable.  If not set, then the data in tvar is replaced.

    Returns
    -------
        None

    Examples
    --------

        >>> pyspedas.store_data('b', data={'x':[2,5,8,11,14,17,20], 'y':[[1,1,1,1,1,1],[2,2,5,4,1,1],[100,100,3,50,1,1],[4,4,8,58,1,1],[5,5,9,21,1,1],[6,6,2,2,1,1],[7,7,1,6,1,1]]})
        >>> pyspedas.derive('b','dbdt')

    """

    a = pyspedas.tplot_tools.data_quants[tvar].differentiate('time')
    if newname is None:
        a.name = tvar
        a.attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)
        pyspedas.tplot_tools.data_quants[tvar] = a
    else:
        data = {'x':a.coords['time'], 'y':a.values * 1e09}  # Internal time in units of nanoseconds
        for coord in a.coords:
            if coord != 'time' and coord != 'spec_bins':
                data[coord] = a.coords[coord].values

        store_data(newname, data=data)
        pyspedas.tplot_tools.data_quants[newname].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[tvar].attrs)

    return



import logging
from pyspedas import (
    wildcard_expand,
    options,
    dailynames,
    download,
    netcdf_to_tplot,
    time_clip as tclip,
)
from .config import CONFIG


def loadr(
    trange=["2023-01-01", "2023-01-02 11:59:59"],
    probe="18",
    instrument="mag",
    datatype="1min",
    prefix="",
    suffix="",
    downloadonly=False,
    no_update=False,
    time_clip=True,
    force_download=False,
):
    """
    Loads GOES-R L2 data (GOES-16 and later) for specified instruments and data types.

    Also loads reprocessed GOES 8-15 data (hires mag and xrs) in GOES-R file format.

    Parameters
    ----------
    trange : list of str
        Time range of interest [starttime, endtime].
    probe : str or int or list of str or int
        GOES spacecraft number(s), e.g., probe=16 or probe=[16, 17].
    instrument : str or list of str
        Name of the instrument (e.g., 'euvs', 'xrs', 'mag', 'mpsh', 'sgps').
    datatype : str, optional
        Data resolution, default is '1min'. Valid options: 'low' (avg), 'hi' (full), among others depending on the instrument.
    prefix : str, optional
        Prefix to add to the tplot variable names. Defaults to an empty string.
    suffix : str, optional
        Suffix to add to the tplot variable names. Defaults to an empty string.
    downloadonly : bool, optional
        If True, downloads the CDF files without loading them into tplot variables. Default is False.
    no_update : bool, optional
        If True, only loads data from the local cache. Default is False.
    time_clip : bool, optional
        If True (the default), clips the variables to exactly the range specified in the trange keyword.
    force_download : bool, optional
        If True, downloads the file even if a newer version exists locally. Default is False.

    Returns
    -------
    list
        List of tplot variables created or list of filenames downloaded.

    Notes
    -----
    - Information can be found at https://www.ngdc.noaa.gov/stp/satellite/goes-r.html
    - GOES-EAST (GOES-16, 2017-2025; GOES-19, 2025-), GOES-WEST (GOES-17, 2018-2022; GOES-18, 2023-).

    Examples
    --------
    >>> from pyspedas.projects.goes import load
    >>> trange = ['2023-01-01', '2023-01-02']
    >>> vars = load(trange=trange, probe='16', instrument='mag', datatype='1min', time_clip=True)
    >>> print(vars)
    ['g16_mag_DQF', 'g16_mag_b_brf', 'g16_mag_b_eci', 'g16_mag_b_epn', 'g16_mag_b_gse',
    'g16_mag_b_gsm', 'g16_mag_b_quality', 'g16_mag_b_total', 'g16_mag_b_vdh',
    'g16_mag_num_points', 'g16_mag_orbit_llr_geo']
    """

    goes16_path_dir = (
        # Data for GOES-R (16 and later)
        "https://data.ngdc.noaa.gov/platforms/solar-space-observing-satellites/goes/"
    )
    goes8_path_dir = (
        # Reprocessed data for GOES 8-15 in GOES-R format
        "https://www.ncei.noaa.gov/data/goes-space-environment-monitor/access/science/"
    )

    time_var = "time"  # name of the time variable in the netcdf files
    out_files = []
    tvars = []

    if isinstance(probe, tuple):
        probe = list(probe)
    if not isinstance(probe, list):
        probe = [probe]
    probe = [str(p) for p in probe]
    if not isinstance(instrument, list):
        instrument = [instrument]

    for prb in probe:
        if int(prb) <= 15:
            goes_path_dir = goes8_path_dir
            remote_path = ""
        else:
            goes_path_dir = goes16_path_dir
            remote_path = "goes" + str(prb) + "/l2/data/"

        pathformat = []

        for instr in instrument:
            if instr == "euvs":
                if int(prb) <= 15:
                    # GOES 8-15
                    logging.warning(
                        "EUVS reprocessed data is not available in reprocessed format for GOES 8-15."
                    )
                    continue

                if datatype in ["full", "hi", "1min", "avg1m"]:  # high resolution 1 min
                    pathformat = [
                        remote_path
                        + "euvs-l2-avg1m_science/%Y/%m/sci_euvs-l2-avg1m_g"
                        + str(prb)
                        + "_d%Y%m%d_v?-?-?.nc"
                    ]
                else:  # low resolution 1 day, smaller files
                    pathformat = [
                        remote_path
                        + "euvs-l2-avg1d_science/%Y/%m/sci_euvs-l2-avg1d_g"
                        + str(prb)
                        + "_d%Y%m%d_v?-?-?.nc"
                    ]
            elif instr == "xrs":
                if datatype in ["full", "hi", "1sec", "flx1s"]:  # high resolution 1 sec
                    if int(prb) <= 15:
                        # GOES 8-15
                        pathformat = [
                            remote_path
                            + "xrs/goes"
                            + str(prb)
                            + "/gxrs-l2-irrad_science/%Y/%m/sci_gxrs-l2-irrad_g"
                            + str(prb)
                            + "_d%Y%m%d_v?-?-?.nc"
                        ]
                    else:
                        pathformat = [
                            remote_path
                            + "xrsf-l2-flx1s_science/%Y/%m/sci_xrsf-l2-flx1s_g"
                            + str(prb)
                            + "_d%Y%m%d_v?-?-?.nc"
                        ]
                else:  # low resolution 1 min, smaller files
                    if int(prb) <= 15:
                        pathformat = [
                            remote_path
                            + "xrs/goes"
                            + str(prb)
                            + "/xrsf-l2-avg1m_science/%Y/%m/sci_xrsf-l2-avg1m_g"
                            + str(prb)
                            + "_d%Y%m%d_v?-?-?.nc"
                        ]
                    else:
                        pathformat = [
                            remote_path
                            + "xrsf-l2-avg1m_science/%Y/%m/sci_xrsf-l2-avg1m_g"
                            + str(prb)
                            + "_d%Y%m%d_v?-?-?.nc"
                        ]
            elif instr == "mag":
                if int(prb) <= 15:
                    # GOES 8-15
                    logging.warning(
                        "Only hires mag data is available in reprocessed format for GOES 8-15."
                    )
                    pathformat = [
                        remote_path
                        + "mag/goes"
                        + str(prb)
                        + "/magn-l2-hires/%Y/%m/dn_magn-l2-hires_g"
                        + str(prb)
                        + "_d%Y%m%d_v?_?_?.nc"  # mag is underscores not hyphens!
                    ]
                else:
                    if datatype in [
                        "full",
                        "hi",
                        "0.1sec",
                        "hires",
                    ]:  # high resolution 0.1 sec
                        pathformat = [
                            remote_path
                            + "magn-l2-hires/%Y/%m/dn_magn-l2-hires_g"
                            + str(prb)
                            + "_d%Y%m%d_v?-?-?.nc"
                        ]
                    else:  # low resolution 1 min, smaller files
                        pathformat = [
                            remote_path
                            + "magn-l2-avg1m/%Y/%m/dn_magn-l2-avg1m_g"
                            + str(prb)
                            + "_d%Y%m%d_v?-?-?.nc"
                        ]
            elif instr == "mpsh":
                if int(prb) <= 15:
                    # GOES 8-15
                    logging.warning(
                        "MPSH reprocessed data is not available in reprocessed format for GOES 8-15."
                    )
                    continue
                time_var = "L2_SciData_TimeStamp"
                if datatype in [
                    "full",
                    "hi",
                    "1min",
                    "avg1m",
                    "1m",
                ]:  # high resolution 1 min
                    pathformat = [
                        remote_path
                        + "mpsh-l2-avg1m/%Y/%m/sci_mpsh-l2-avg1m_g"
                        + str(prb)
                        + "_d%Y%m%d_v?-?-?.nc"
                    ]
                else:  # low resolution 5 min, smaller files
                    pathformat = [
                        remote_path
                        + "mpsh-l2-avg5m/%Y/%m/sci_mpsh-l2-avg5m_g"
                        + str(prb)
                        + "_d%Y%m%d_v?-?-?.nc"
                    ]
            elif instr == "sgps":
                if int(prb) <= 15:
                    # GOES 8-15
                    logging.warning(
                        "SGPS reprocessed data is not available in reprocessed format for GOES 8-15."
                    )
                    continue
                if datatype in [
                    "full",
                    "hi",
                    "1min",
                    "avg1m",
                    "1m",
                ]:  # high resolution 1 min
                    pathformat = [
                        remote_path
                        + "sgps-l2-avg1m/%Y/%m/sci_sgps-l2-avg1m_g"
                        + str(prb)
                        + "_d%Y%m%d_v?-?-?.nc"
                    ]
                else:  # low resolution 5 min, smaller files
                    pathformat = [
                        remote_path
                        + "sgps-l2-avg5m/%Y/%m/sci_sgps-l2-avg5m_g"
                        + str(prb)
                        + "_d%Y%m%d_v?-?-?.nc"
                    ]

            # find the full remote path names using the trange
            if not isinstance(pathformat, list):
                pathformat = [pathformat]

            remote_names = []
            for path in pathformat:
                remote_names.extend(dailynames(file_format=path, trange=trange))

            # For each probe and instrument, download the files
            files = download(
                remote_file=remote_names,
                remote_path=goes_path_dir,
                local_path=CONFIG["local_data_dir"],
                no_download=no_update,
                force_download=force_download,
            )

            files = sorted(set(files))
            out_files.extend(files)

            if len(files) > 0 and not downloadonly:
                if prefix is None or prefix == "" or prefix == "probename":
                    # Example of prefix: g16_xrs_
                    prefix_local = "g" + str(prb) + "_" + instr + "_"
                else:
                    prefix_local = prefix

                # Load all files from the same probe and instrument, merging them
                tvars_local = netcdf_to_tplot(
                    files, prefix=prefix_local, suffix=suffix, time=time_var
                )
                if len(tvars_local) > 0:
                    tvars.extend(tvars_local)

    if time_clip:
        tclip(tvars, trange[0], trange[1], overwrite=True)

    if downloadonly:
        return out_files

    tvars = sorted(set(tvars))
    return tvars


def load(
    trange=["2013-11-05", "2013-11-06 12:00:00"],
    probe="15",
    instrument="fgm",
    datatype="1min",
    prefix="",
    suffix="",
    downloadonly=False,
    no_update=False,
    time_clip=True,
    force_download=False,
    goes_r=False,
):
    """
    Load GOES L2 data.

    Parameters
    ----------
    trange : list of str
        Time range of interest ['YYYY-MM-DD', 'YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss', 'YYYY-MM-DD/hh:mm:ss'].
    probe : str, int, or list of str/int
        GOES spacecraft number(s), e.g., probe=15.
    instrument : str or list of str
        Name of the instrument. For GOES 8-15: 'fgm', 'eps', 'epead', 'maged', 'magpd', 'hepad', 'xrs'.
        For GOES-R 16-18: 'euvs', 'xrs', 'mag', 'mpsh', 'sgps'.
    datatype : str
        Data type; usually instrument resolution, depends on the instrument (default '1min').
        Valid for GOES 8-15: 'hi', 'low', 'full', 'avg', '1min', '5min'.
        Valid for GOES-R 16-18: 'hi', 'low', 'full', 'avg', and other options.
    prefix : str, optional
        Prefix to add to the tplot variable names. By default, no prefix is added.
        If prefix is 'probename', then the name will be used, for example, 'g16'.
    suffix : str, optional
        Suffix to add to the tplot variable names. By default, no suffix is added.
    downloadonly : bool, optional
        If True, downloads the CDF files without loading them into tplot variables. Default is False.
    no_update : bool, optional
        If True, only loads data from the local cache. Default is False.
    time_clip : bool, optional
        If True (the default), clips the variables to exactly the range specified in the trange keyword.
    force_download : bool, optional
        If True, downloads the file even if a newer version exists locally. Default is False.
    goes_r : bool, optional
        If True, loads reprocessed data in GOES-R file format, for probes 8-15.
        Default is False, loads original data for probes 8-15.

    Returns
    -------
    list of str
        List of tplot variables created or list of filenames downloaded.

    Notes
    -----
    This function can be used to load data from three sets of GOES data:

    1. GOES 1-15, original netcdf format:
    https://www.ncei.noaa.gov/data/goes-space-environment-monitor/access/avg/

    2. GOES-R 16-19, revised netcdf format:
    https://data.ngdc.noaa.gov/platforms/solar-space-observing-satellites/goes/

    3. GOES 8-15, reprocessed data, GOES-R (revised) netcdf format (hires mag, and xrs only):
    for this, use the goes_r=True option
    https://www.ncei.noaa.gov/data/goes-space-environment-monitor/access/science/


    Examples
    --------
    >>> from pyspedas.projects.goes import load
    >>> trange = ['2019-01-01', '2019-01-02']
    >>> vars = load(trange=trange, probe='15', instrument='fgm', datatype='1min', time_clip=True)
    >>> print(vars)
    ['g15_fgm_BX_1_QUAL_FLAG', 'g15_fgm_BX_1_NUM_PTS', 'g15_fgm_BX_1', ...]
    (the result is a list of 66 variables)
    """

    if isinstance(probe, tuple):
        probe = list(probe)
    if not isinstance(probe, list):
        probe = [probe]
    probe = [str(p) for p in probe]
    if not isinstance(instrument, list):
        instrument = [instrument]

    probe_r = []  # GOES-R probes
    probe_s = []  # GOES probes
    out_files_r = []  # GOES-R files to download
    out_files = []  # GOES files to download
    tvars_r = []  # GOES-R list of variables created
    tvars = []  # all variables created

    # Find if we need to call the GOES-R load function for some probes
    for prb in probe:
        if int(prb) > 15 or goes_r:
            probe_r.append(str(prb))
        else:
            probe_s.append(str(prb))

    if len(probe_r):
        tvars_r = loadr(
            trange=trange,
            probe=probe_r,
            instrument=instrument,
            datatype=datatype,
            prefix=prefix,
            suffix=suffix,
            downloadonly=downloadonly,
            no_update=no_update,
            time_clip=time_clip,
            force_download=force_download,
        )
        if downloadonly:
            out_files_r = tvars_r

    # Continue with loading GOES (1-15) data
    for prb in probe_s:
        avg_path = "avg/%Y/%m/goes" + str(prb) + "/netcdf/" + "g" + str(prb)
        full_path = "full/%Y/%m/goes" + str(prb) + "/netcdf/" + "g" + str(prb)

        for instr in instrument:
            pathformat = []
            if instr == "fgm":
                if datatype == "512ms" or datatype == "full":  # full, unaveraged data
                    pathformat = full_path + "_magneto_512ms_%Y%m%d_%Y%m%d.nc"
                elif datatype == "5min":  # 5 min averages
                    pathformat = avg_path + "_magneto_5m_%Y%m01_%Y%m??.nc"
                else:  # 1 min averages, goes13, goes15 only contain 1m averages
                    pathformat = avg_path + "_magneto_1m_%Y%m01_%Y%m??.nc"
            elif instr == "eps":
                # energetic particle sensor -- only valid for GOES-08 through GOES-12, only averaged data available
                if datatype == "1min" or datatype == "full":
                    pathformat = avg_path + "_eps_1m_%Y%m01_%Y%m??.nc"
                else:  # 'low' or 5min
                    pathformat = avg_path + "_eps_5m_%Y%m01_%Y%m??.nc"
            elif instr == "epead":
                # electron, proton, alpha detector -- only valid on GOES-13, 14, 15
                if datatype == "1min":
                    pathformat = [
                        "_epead_e13ew_1m_%Y%m01_%Y%m??.nc",
                        "_epead_p17ew_1m_%Y%m01_%Y%m??.nc",
                        "_epead_a16ew_1m_%Y%m01_%Y%m??.nc",
                    ]
                    pathformat = [avg_path + s for s in pathformat]
                elif datatype == "5min" or datatype == "low":
                    pathformat = [
                        "_epead_e13ew_5m_%Y%m01_%Y%m??.nc",
                        "_epead_p17ew_5m_%Y%m01_%Y%m??.nc",
                        "_epead_a16ew_5m_%Y%m01_%Y%m??.nc",
                    ]
                    pathformat = [avg_path + s for s in pathformat]
                else:  # full
                    pathformat = [
                        "_epead_a16e_32s_%Y%m%d_%Y%m%d.nc",
                        "_epead_a16w_32s_%Y%m%d_%Y%m%d.nc",
                        "_epead_e1ew_4s_%Y%m%d_%Y%m%d.nc",
                        "_epead_e2ew_16s_%Y%m%d_%Y%m%d.nc",
                        "_epead_e3ew_16s_%Y%m%d_%Y%m%d.nc",
                        "_epead_p1ew_8s_%Y%m%d_%Y%m%d.nc",
                        "_epead_p27e_32s_%Y%m%d_%Y%m%d.nc",
                        "_epead_p27w_32s_%Y%m%d_%Y%m%d.nc",
                    ]
                    pathformat = [full_path + s for s in pathformat]
            elif instr == "maged":
                # magnetospheric electron detector -- only valid on GOES 13, 14, 15
                if datatype == "1min":
                    pathformat = avg_path + "_maged_19me15_1m_%Y%m01_%Y%m??.nc"
                elif datatype == "5min" or datatype == "low":
                    pathformat = avg_path + "_maged_19me15_5m_%Y%m01_%Y%m??.nc"
                else:  # full
                    channels = ["me1", "me2", "me3", "me4", "me5"]
                    resolution = ["2", "2", "4", "16", "32"]
                    pathformat = []
                    for idx, channel in enumerate(channels):
                        pathformat.append(
                            "_maged_19"
                            + channel
                            + "_"
                            + resolution[idx]
                            + "s_%Y%m%d_%Y%m%d.nc"
                        )
                    pathformat = [full_path + s for s in pathformat]
            elif instr == "magpd":
                # magnetospheric proton detector -- only valid on GOES 13, 14, 15
                if datatype == "1min" or datatype == "low":
                    pathformat = avg_path + "_magpd_19mp15_1m_%Y%m01_%Y%m??.nc"
                else:  # full
                    channels = ["mp1", "mp2", "mp3", "mp4", "mp5"]
                    resolution = ["16", "16", "16", "32", "32"]
                    pathformat = []
                    for idx, channel in enumerate(channels):
                        pathformat.append(
                            "_magpd_19"
                            + channel
                            + "_"
                            + resolution[idx]
                            + "s_%Y%m%d_%Y%m%d.nc"
                        )
                    pathformat = [full_path + s for s in pathformat]
            elif instr == "hepad":
                # high energy proton and alpha detector -- valid for GOES 08-15
                if datatype == "1min":
                    pathformat = [
                        "_hepad_ap_1m_%Y%m01_%Y%m??.nc",
                        "_hepad_s15_1m_%Y%m01_%Y%m??.nc",
                    ]
                    pathformat = [avg_path + s for s in pathformat]
                elif datatype == "5min" or datatype == "low":
                    pathformat = [
                        "_hepad_ap_5m_%Y%m01_%Y%m??.nc",
                        "_hepad_s15_5m_%Y%m01_%Y%m??.nc",
                    ]
                    pathformat = [avg_path + s for s in pathformat]
                else:
                    pathformat = [
                        "_hepad_ap_32s_%Y%m%d_%Y%m%d.nc",
                        "_hepad_s15_4s_%Y%m%d_%Y%m%d.nc",
                    ]
                    pathformat = [full_path + s for s in pathformat]
            elif instr == "xrs":
                # x-ray sensor -- valid for GOES 08-15
                if datatype == "1min":
                    pathformat = avg_path + "_xrs_1m_%Y%m01_%Y%m??.nc"
                elif datatype == "5min" or datatype == "low":
                    pathformat = avg_path + "_xrs_5m_%Y%m01_%Y%m??.nc"
                else:
                    pathformat = [
                        "_xrs_2s_%Y%m%d_%Y%m%d.nc",
                        "_xrs_3s_%Y%m%d_%Y%m%d.nc",
                    ]
                    pathformat = [full_path + s for s in pathformat]

            # find the full remote path names using the trange
            if not isinstance(pathformat, list):
                pathformat = [pathformat]

            remote_names = []
            for path in pathformat:
                remote_names.extend(dailynames(file_format=path, trange=trange))

            files = download(
                remote_file=remote_names,
                remote_path=CONFIG["remote_data_dir"],
                local_path=CONFIG["local_data_dir"],
                no_download=no_update,
                force_download=force_download,
            )

            if files is not None:
                if not isinstance(files, list):
                    files = [files]
                out_files.extend(files)

            tvars_local = []
            if len(files) > 0 and downloadonly is False:
                if prefix is None or prefix == "" or prefix == "probename":
                    # Example of prefix: g15_xrs_
                    prefix_local = "g" + str(prb) + "_" + instr + "_"
                else:
                    prefix_local = prefix

                tvars_local = netcdf_to_tplot(
                    files, prefix=prefix_local, suffix=suffix, time="time_tag"
                )

                if len(tvars_local) > 0:
                    tvars.extend(tvars_local)

    if time_clip:
        tclip(tvars, trange[0], trange[1], suffix="", overwrite=True)

    if downloadonly:
        out_files.extend(out_files_r)  # append GOES-R filenames
        out_files = sorted(set(out_files))
        return out_files

    if len(tvars_r):
        tvars.extend(tvars_r)  # append GOES-R variables

    # This might make sense in the XRS load routine, but then we wouldn't be able to use partial_function for it...
    if "xrs" in instrument:
        xray_flux_vars = wildcard_expand(tvars, "*_xrs_*_flux", quiet=True)
        if len(xray_flux_vars) > 0:
            options(xray_flux_vars, "ylog", True)
    return tvars



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyspedas
import logging

def get_timespan(name=None):
    """
    This function extracts the time span from the Tplot Variables stored in memory.
    If called with no arguments, return the global timespan set by timespan(), or None if none has been set.
    
    Parameters
    ----------
        name : str 
            Name of the tplot variable
         
    Returns
    -------
    list of float
        time_begin : float
            The beginning of the time series
        time_end : float
            The end of the time series
            
    Examples
    --------
        >>> # Retrieve the time span from Variable 1
        >>> import pyspedas
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store_data("Variable1", data={'x':x_data, 'y':y_data})
        >>> time1, time2 = pyspedas.get_timespan("Variable1")

    """

    if name is None:
        if 'x_range' in pyspedas.tplot_tools.tplot_opt_glob.keys():
            return pyspedas.tplot_tools.tplot_opt_glob['x_range']
        else:
            return None
    elif name not in pyspedas.tplot_tools.data_quants.keys():
        logging.info("The name %s is currently not in pyspedas",name)
        return None

    return pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['trange'][0], pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['trange'][1]


import numpy as np
import logging
from pyspedas.tplot_tools import tnames, store_data, time_double, degap, tdeflag, time_clip
from pyspedas.projects.omni.load import load as omni_load
from pyspedas.utilities.tcopy import tcopy_one


def get_omni_default_solarwind(trange):
    """ Get some default values for solar wind data."""
    times = time_double(trange[0])
    timee = time_double(trange[1])

    nnom = 1000
    dt = (timee - times) / nnom
    tgrid = times + dt * np.arange(nnom)

    # Create default pressure and Bz arrays
    bzout = np.zeros(nnom)
    dpout = np.full(nnom, 2.088)  # nPa

    return [tgrid, dpout, bzout]


def omni_solarwind_load(trange, prefix=None, suffix=None, level="hro", min5=False):
    """
    Loads solar wind data from OMNI.

    Parameters
    ----------
    trange : list of float
        Time range of data to be loaded.
    level : str, optional
        Data level for OMNI data; valid options: 'hro', 'hro2'.
        Default is 'hro2'.
    min5 : bool, optional
        Flag indicating whether to load 1-minute or 5-minute data. Default is False (1-minute data).

    Returns
    -------
    list of str
        List of tplot variable names created for BZ and P.

    """

    if min5:
        datatype = "5min"
    else:
        datatype = "1min"

    if prefix is None:
        prefix = "omni_solarwind_"
    if suffix is None:
        suffix = ""

    if level is None or level not in ["hro", "hro2"]:
        level = "hro2"

    new_bz = prefix + "BZ" + suffix
    new_p = prefix + "P" + suffix

    omni_load(trange=trange, level=level, datatype=datatype)

    bz_name = tnames("BZ_GSM")
    p_name = tnames("Pressure")
    if bz_name is None or p_name is None or len(bz_name) < 1 or len(p_name) < 1:
        logging.warning("OMNI Solar Wind data not found. Loading default values.")
        defd = get_omni_default_solarwind(trange)
        tout = defd[0]
        dpout = defd[1]
        bzout = defd[2]
        store_data(new_p, data={"x": tout, "y": dpout})
        store_data(new_bz, data={"x": tout, "y": bzout})
    else:
        tcopy_one(bz_name[0], new_bz)
        tcopy_one(p_name[0], new_p)
   
    degap(new_bz)
    degap(new_p)
    tdeflag([new_bz, new_p], method="remove_nan", overwrite=True)
    time_clip([new_bz, new_p], trange[0], trange[1], overwrite=True)

    return [new_bz, new_p]



import logging
import requests
from pyspedas.projects.themis.load import load


class Themis_gmag:
    """This object holds the dictionary of gmag stations and their metadata"""

    def __init__(self):
        self.gmag_dict = {}

    def query_gmags(self):
        """Fills a dictionary of gmag stations and all their metadata"""

        url = "http://themis.ssl.berkeley.edu/gmag/gmag_json.php"

        params = dict(station="", group="")
        resp = requests.get(url=url, params=params)
        data = resp.json()
        self.gmag_dict = data

    def get_gmag_list(self):
        """returns the gmag dictionary, querying if necessary"""
        if self.gmag_dict == {}:
            self.query_gmags()
        return self.gmag_dict


themis_gmag_dict = Themis_gmag()


def gmag(
    trange=["2007-03-23", "2007-03-24"],
    sites=None,
    group=None,
    level="l2",
    prefix="",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Load ground magnetometer data from the THEMIS mission.

    Parameters
    ----------
    trange : list, optional
        Time range of interest in the format ["start_date", "end_date"].
        Default is ["2007-03-23", "2007-03-24"].
    sites : list or str, optional
        List of ground magnetometer sites to load data for. If None, data
        for all available sites will be loaded. Default is None.
    group : str, optional
        Name of a pre-defined group of sites to load data for. If specified,
        the 'sites' parameter will be ignored. Default is None.
    level : str, optional
        Data level to load. Default is "l2".
    prefix : str, optional
        Prefix to append to the variable names. Default is "".
    suffix : str, optional
        Suffix to append to the variable names. Default is "".
    get_support_data : bool, optional
        Flag indicating whether to download support data. Default is False.
    varformat : str, optional
        Format of the variable names. Default is None.
    varnames : list, optional
        List of specific variable names to load. Default is [].
    downloadonly : bool, optional
        Flag indicating whether to only download the data without loading it.
        Default is False.
    notplot : bool, optional
        Flag indicating whether to plot the loaded data. Default is False.
    no_update : bool, optional
        Flag indicating whether to update the data files. Default is False.
    time_clip : bool, optional
        Flag indicating whether to clip the data to the specified time range.
        Default is False.
    force_download: bool, optional
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------
    dict
        A dictionary containing the loaded data.

    Examples
    --------
    >>> from pyspedas.projects.themis import gmag
    >>> from pyspedas import tplot
    >>>
    >>> # Load ground magnetometer data for specific sites and time range
    >>> gmag_vars = gmag(sites=['ccnv','bmls'], trange=['2013-11-05', '2013-11-06'])
    >>> tplot(['thg_mag_bmls', 'thg_mag_ccnv'])
    """

    if sites is None:
        thm_sites = (
            "idx atha chbg ekat fsim fsmi fykn gbay glyn gill inuv kapu "
            "kian kuuj mcgr nrsq pgeo pina rank snap snkq tpas whit "
            "yknf".split(" ")
        )
        tgo_sites = [
            "nal",
            "lyr",
            "hop",
            "bjn",
            "nor",
            "sor",
            "tro",
            "and",
            "don",
            "rvk",
            "sol",
            "kar",
            "jan",
            "jck",
            "dob",
        ]
        dtu_sites = [
            "atu",
            "dmh",
            "svs",
            "tdc",
            "bfe",
            "roe",
            "thl",
            "kuv",
            "upn",
            "umq",
            "gdh",
            "stf",
            "skt",
            "ghb",
            "fhb",
            "naq",
            "amk",
            "sco",
            "tab",
            "sum",
            "hov",
        ]
        ua_sites = [
            "arct",
            "bett",
            "cigo",
            "eagl",
            "fykn",
            "gako",
            "hlms",
            "homr",
            "kako",
            "pokr",
            "trap",
        ]
        maccs_sites = ["cdrt", "chbr", "crvr", "gjoa", "iglo", "nain", "pang", "rbay"]
        usgs_sites = [
            "bou",
            "brw",
            "bsl",
            "cmo",
            "ded",
            "frd",
            "frn",
            "gua",
            "hon",
            "new",
            "shu",
            "sit",
            "sjg",
            "tuc",
        ]
        atha_sites = [
            "roth",
            "leth",
            "redr",
            "larg",
            "vldr",
            "salu",
            "akul",
            "puvr",
            "inuk",
            "kjpk",
            "radi",
            "stfl",
            "sept",
            "schf",
        ]
        epo_sites = [
            "bmls",
            "ccnv",
            "drby",
            "fyts",
            "hots",
            "loys",
            "pgeo",
            "pine",
            "ptrs",
            "rmus",
            "swno",
            "ukia",
        ]
        falcon_sites = ["hris", "kodk", "lrel", "pblo", "stfd", "wlps"]
        mcmac_sites = ["amer", "benn", "glyn", "lyfd", "pcel", "rich", "satx", "wrth"]
        nrcan_sites = ["blc", "cbb", "iqa", "mea", "ott", "stj", "vic"]
        step_sites = ["fsj", "ftn", "hrp", "lcl", "lrg", "pks", "whs"]
        fmi_sites = [
            "han",
            "iva",
            "kev",
            "kil",
            "mas",
            "mek",
            "muo",
            "nur",
            "ouj",
            "pel",
            "ran",
            "tar",
        ]
        aair_sites = ["amd", "bbg", "brn", "dik", "loz", "pbk", "tik", "viz"]
        carisma_sites = [
            "anna",
            "back",
            "cont",
            "daws",
            "eski",
            "fchp",
            "fchu",
            "gull",
            "isll",
            "lgrr",
            "mcmu",
            "mstk",
            "norm",
            "osak",
            "oxfo",
            "pols",
            "rabb",
            "sach",
            "talo",
            "thrf",
            "vulc",
            "weyb",
            "wgry",
        ]
        sites = (
            thm_sites
            + tgo_sites
            + dtu_sites
            + ua_sites
            + maccs_sites
            + usgs_sites
            + atha_sites
            + epo_sites
            + falcon_sites
            + mcmac_sites
            + nrcan_sites
            + step_sites
            + fmi_sites
            + aair_sites
            + carisma_sites
        )

    if group is not None:
        sites = eval(group + "_sites")

    if not isinstance(sites, list):
        sites = [sites]

    # check for sites in Greenland
    greenland = []
    for site in sites:
        if check_greenland(site):
            greenland.append(True)
        else:
            greenland.append(False)

    return load(
        instrument="gmag",
        trange=trange,
        level=level,
        prefix=prefix,
        suffix=suffix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        downloadonly=downloadonly,
        notplot=notplot,
        stations=sites,
        greenland=greenland,
        time_clip=time_clip,
        no_update=no_update,
        force_download=force_download,
    )


def get_group(station_name):
    """
    Returns a list of groups a particular station belongs to.

    Parameters
    ----------
    station_name : str, required
        The station name.

    Returns
    -------
    list
        A list of station groups.

    Examples
    --------
    >>> from pyspedas.projects.themis.ground.gmag import get_group
    >>> atha_groups = get_group("atha")
    >>> print(atha_groups)
    ['ae', 'asi', 'carisma']
    """

    group_list = []
    gmag_dict = themis_gmag_dict.get_gmag_list()
    station_name = station_name.lower()

    for station in gmag_dict:
        if station["ccode"].lower() == station_name:
            for key, value in station.items():
                if value == "Y":
                    if station["ccode"].lower() not in group_list:
                        group_list.append(key)

    return group_list


def gmag_list(group="all"):
    """
    Returns a list of ground magnetometer station names based on the specified group.

    Parameters
    ----------
    group : str, optional
        The group of stations to filter by. Default is "all".

    Returns
    -------
    list
        A list of station names.

    Examples
    --------
    >>> from pyspedas.projects.themis.ground.gmag import gmag_list
    >>> epo_list = gmag_list("EPO")
    >>> print(epo_list)
    ['bmls', 'ccnv', 'drby', 'hots', 'loys', 'ptrs', 'ukia']
    """

    station_list = []
    gmag_dict = themis_gmag_dict.get_gmag_list()
    group = group.lower()

    for station in gmag_dict:
        station_name = station["ccode"].lower()
        station_group = get_group(station_name)
        if group in ["all", "*", ""] or group in station_group:
            station_list.append(station_name)
            logging.info(
                station_name
                + ": from "
                + station["day_first"]
                + " to "
                + station["day_last"]
            )

    return station_list


def gmag_groups():
    """
    Returns a dictionary containing groups of stations and their station codes.

    Returns
    -------
    dict
        A dictionary where the keys represent the gmag values and the values are lists of station codes.

    Examples
    --------
    >>> from pyspedas.projects.themis.ground.gmag import gmag_groups
    >>> groups = gmag_groups()
    >>> groups.keys()  # all groups
    dict_keys(['kyoto', 'sgu', 'autx', 'ae', 'aari', 'dtu', 'greenland', 'tgo', 'asi', 'carisma', 'mcmac',
    'nrcan', 'epo', 'usgs', 'maccs', 'gima', 'cgsm', 'gbo', 'autu', 'leirv', 'pen', 'falcon', 'fmi', 'secs', 'bas'])
    >>> groups['epo']  # stations in the EPO group
    ['bmls', 'ccnv', 'drby', 'hots', 'loys', 'ptrs', 'ukia']
    """
    group_dict = {}
    gmag_dict = themis_gmag_dict.get_gmag_list()

    for station in gmag_dict:
        for key, value in station.items():
            if value == "Y" or value == "N":
                if key in group_dict:
                    if station["ccode"].lower() not in group_dict[key]:
                        group_dict[key].append(station["ccode"].lower())
                else:
                    group_dict[key] = []
                    group_dict[key].append(station["ccode"].lower())

    # print the groups
    for g, s in group_dict.items():
        logging.info(g + ":" + ",'".join(s) + "'")

    return group_dict


def check_gmag(station_name):
    """
    Check if a station name exists in the gmag dictionary.

    Parameters
    ----------
    station_name : str
        The name of the station to check.

    Returns
    -------
    int
        Returns 1 if the station name exists in the gmag dictionary, otherwise returns 0.

    Examples
    --------
    >>> from pyspedas.projects.themis.ground.gmag import check_gmag
    >>> check_gmag("atha")
    1
    """

    gmag_dict = themis_gmag_dict.get_gmag_list()

    for station in gmag_dict:
        if station["ccode"].lower() == station_name.lower():
            return 1

    return 0


def check_greenland(station_name):
    """
    Check if a given station belongs to the Greenland group.

    Parameters
    ----------
    station_name : str
        The name of the station to check.

    Returns
    -------
    int
        Returns 1 if the station is in Greenland, otherwise returns 0.

    Examples
    --------
    >>> from pyspedas.projects.themis.ground.gmag import check_greenland
    >>> check_greenland("bfe")
    1
    """

    gmag_dict = themis_gmag_dict.get_gmag_list()

    for station in gmag_dict:
        if station["ccode"].lower() == station_name.lower():
            if (
                station["country"] is not None
                and station["country"].lower() == "greenland"
            ) or (
                station["greenland"] is not None and station["greenland"].lower() == "y"
            ):
                return 1

    return 0



"""
This is a python version of the wavelet.pro script in IDL SPEDAS.
It computes the wavelet transform as described by Torrence & Compo (1998).

This function is a wrapper for wavelet98.py with some changes:
    - It can handle 1D or 2D data (while wavelet98.py can only handle 1D).
    - It uses the Morlet wavelet only (while wavelet98.py can also use Paul or DOG wavelets).
    - The default Morlet parameter is w0=2pi (while the wavelet98.py default is 6.0).
    - The default dj is dj=1/8 * (2pi/w0) (which is 1/8., the same as in wavelet98.py when w0=2pi).
    - Computes the scales.
    - Uses the normalization factor np.sqrt(2 * dt).
"""

import numpy as np
from pyspedas.analysis.wavelet98 import wavelet98


def wavelet2(
    y,
    dt,
    prange=None,
    frange=None,
    pad=False,
    period=None,
    dj=None,
    param=None,
):
    """
    Wrapper for the Torrence & Compo wavelet98.py function.
    Uses Morlet mother wavelet.

    Parameters
    ----------
    y : ndarray
        Input data, 1D or 2D (n x d2)
    dt : float
        Time step
    prange : list or tuple, optional
        Period range [min, max]
    frange : list or tuple, optional
        Frequency range [min, max] (overrides prange if set)
    pad : bool, optional
        If True, pad the time series for FFT
    period : array, optional
        Output: periods used (filled by wavelet)
    dj : float, optional
        Scale spacing
    param : float, optional
        Morlet parameter (w0)

    Returns
    -------
    wave : ndarray
        Wavelet transform (n, jv+1, d2)
        (n = number of time points, jv = number of scales, d2 = 1 or 2)
    period : ndarray
        Periods corresponding to each scale (jv+1)
    """
    y = np.asarray(y)
    if y.ndim == 1:
        y = y[:, np.newaxis]
    n, d2 = y.shape

    # Set default Morlet parameter if not provided
    if param is None:
        w0 = 2.0 * np.pi
    else:
        w0 = param

    # Set default dj if not provided
    if dj is None:
        dj = 1.0 / 8.0 * (2.0 * np.pi / w0)

    # If frange is set, override prange
    if frange is not None:
        prange = [min(1.0 / np.array(frange)), max(1.0 / np.array(frange))]
    # Default prange: [nyquist period, 5% of total interval]
    if prange is None:
        prange = [2.0 * dt, 0.05 * n * dt]
    if prange[0] == 0:
        prange[0] = 2.0 * dt

    # Compute srange (scales) from prange and Morlet factor
    morlet_factor = (w0 + np.sqrt(2 + w0**2)) / (4 * np.pi)
    srange = np.array(prange) * morlet_factor

    # Number of scales (jv)
    jv = int(np.floor(np.log(srange[1] / srange[0]) / np.log(2) / dj))
    if jv <= 0:
        return -1, None

    # Output arrays
    wave = np.zeros((n, jv + 1, d2), dtype=complex)
    period_arr = None

    # Loop over each column (component)
    for d in range(d2):
        # Call the main wavelet function for each component
        wv98 = wavelet98(
            y[:, d], dt, s0=srange[0], dj=dj, j_scales=jv, pad=pad, param=w0
        )
        wv = wv98[0]
        period = wv98[2]
        wave[:, :, d] = wv
        # IDL multiplies by sqrt(2*dt) for normalization
        wave[:, :, d] *= np.sqrt(2 * dt)
        if period_arr is None:
            period_arr = period

    # Remove singleton dimensions
    if wave.ndim > 0 and wave.shape[-1] == 1:
        wave = wave.squeeze(axis=-1)

    return wave, period_arr



import os

CONFIG = {'local_data_dir': 'mth5/',
          'remote_data_dir': ''}  # Reserved for the future use

# override local data directory with environment variables
if os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir'] = os.sep.join([os.environ['SPEDAS_DATA_DIR'],
                                           'mth5'])


import os
import unittest
from pyspedas.tplot_tools import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_cospin_data(self):
        data = pyspedas.projects.ulysses.cospin()
        self.assertTrue(data_exists('Protons'))
        self.assertTrue(data_exists('Electrons'))
        self.assertTrue(data_exists('HiE_protons'))
        self.assertTrue(data_exists('Z_ge_3'))

    def test_load_vhm_data(self):
        data = pyspedas.projects.ulysses.vhm()
        self.assertTrue(data_exists('B_MAG'))
        data = pyspedas.projects.ulysses.vhm(notplot=True)
        self.assertTrue('B_MAG' in data)

    def test_load_swoops_data(self):
        data = pyspedas.projects.ulysses.swoops()
        self.assertTrue(data_exists('Density'))
        self.assertTrue(data_exists('Temperature'))
        self.assertTrue(data_exists('Velocity'))
        data = pyspedas.projects.ulysses.swoops(datatype='proton-moments_swoops')
        self.assertTrue(data_exists('Tpar'))
        self.assertTrue(data_exists('Tper'))
        self.assertTrue(data_exists('dens'))

    def test_load_swics_data(self):
        data = pyspedas.projects.ulysses.swics()
        self.assertTrue(data_exists('Velocity'))

    def test_load_epac_data(self):
        data = pyspedas.projects.ulysses.epac()
        self.assertTrue(data_exists('Omni_Protons'))

    def test_load_hiscale_data(self):
        data = pyspedas.projects.ulysses.hiscale()
        self.assertTrue(data_exists('Electrons'))
        
    def test_load_grb_data(self):
        data = pyspedas.projects.ulysses.grb()
        self.assertTrue(data_exists('Count_Rate'))

    def test_load_grb_data_prefix_none(self):
        data = pyspedas.projects.ulysses.grb(prefix=None)
        self.assertTrue(data_exists('Count_Rate'))

    def test_load_grb_data_suffix_none(self):
        data = pyspedas.projects.ulysses.grb(suffix=None)
        self.assertTrue(data_exists('Count_Rate'))

    def test_load_grb_data_prefix_suffix(self):
        data = pyspedas.projects.ulysses.grb(prefix='pre_', suffix='_suf')
        self.assertTrue(data_exists('pre_Count_Rate_suf'))

    def test_downloadonly(self):
        files = pyspedas.projects.ulysses.urap(downloadonly=True, trange=['2003-01-01', '2003-01-02'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



from pyspedas.projects.mms.mms_events import mms_brst_events

import unittest


class TestMMSBrstEvents(unittest.TestCase):
    def test_trange(self):
        # Test with trange option
        trange = ['2015-10-16', '2015-10-17']
        mms_brst_events(trange=trange)

    def test_reload(self):
        # Test with reload option
        reload = True
        mms_brst_events(reload=reload)


if __name__ == '__main__':
    unittest.main()
